Add LoadPath projectPath.

Require Import relation_definitions.
Require Import propositional_formulas.
Require Import List.
Require Import Bool.
Require Import Arith.
Require Import Omega.

(*----------genBES----------*)

Inductive booleanEquation := bEqn : propVar -> propForm -> booleanEquation.

Lemma eq_dec_bEqn : eq_dec booleanEquation.
Proof.
intro;intros.
induction x;induction y;auto.
destruct (eq_dec_nat p p1).
rewrite e in *.

destruct (eq_dec_propForm p0 p2).
rewrite e0 in *;left;auto.
right;intro.
apply n.
congruence.
right;intro.
apply n.
congruence.
Qed.

Record block:=makeBlock{
  theBlock:> list booleanEquation;
  non_empty: exists beqn:booleanEquation, In beqn theBlock; (*blocks are non-empty*)
  blockType: bool
}.

Fixpoint bnd_block (E:list booleanEquation)(x:propVar):Prop:=
match E with 
| nil => False
| cons (bEqn y _) E' => x=y \/ bnd_block E' x
end.

Fixpoint bnd (E:list block)(x:propVar) : Prop :=
match E with
| nil => False
| cons bl E' => bnd_block bl x \/ bnd E' x
end.

Fixpoint bnd_cnt_block (bl:list booleanEquation)(x:propVar): nat :=
match bl with
| nil => 0
| bEqn y _ :: bl' => if (beq_nat x y) then S(bnd_cnt_block bl' x) else bnd_cnt_block bl' x
end.

Fixpoint bnd_cnt (E:list block)(x:propVar) : nat :=
match E with
| nil => 0
| cons bl E' => bnd_cnt_block bl x + bnd_cnt E' x
end.

Lemma bnd_cnt_block_gt : forall bl:list booleanEquation, forall x :propVar, bnd_block bl x <-> bnd_cnt_block bl x > 0.
Proof.
induction bl.
split;intro.
inversion H.
simpl in H.
inversion H.

destruct a as [z fz].
simpl in *.
split;intro.
destruct H.
rewrite H.
rewrite<- beq_nat_refl.
omega.
destruct (IHbl x).
specialize (H0 H).
destruct (beq_nat x z);omega.

destruct (eq_nat_decide x z).
rewrite (eq_nat_eq _ _ e);auto.
right.
apply IHbl.
assert (beq_nat x z=false).
  apply beq_nat_false_iff.
  intro.
  apply n;apply eq_nat_is_eq;auto.
rewrite H0 in H.
assumption.
Qed.

Lemma bnd_cnt_gt : forall bl:list block, forall x: propVar, bnd bl x <-> bnd_cnt bl x > 0.
Proof.
induction bl.
split;intro.
inversion H.
simpl in H.
inversion H.

destruct a as [b].
simpl in *.
split;intro.
destruct H.
assert (bnd_cnt_block b x>0).
  apply bnd_cnt_block_gt;auto.
omega.
destruct (IHbl x).
specialize (H0 H);omega.
destruct (gt_dec (bnd_cnt bl x) 0).
right;apply IHbl;auto.
assert (bnd_cnt_block b x>0).
  omega.
left;apply bnd_cnt_block_gt;auto.
Qed.

Lemma bnd_block_dec : forall l:list booleanEquation, forall x:propVar, 
{bnd_block l x} + {~bnd_block l x}.
Proof.
induction l;simpl;intros;auto.
destruct a as [y f];simpl in *.
destruct (eq_nat_decide x y).
rewrite (eq_nat_eq _ _ e) in *;auto.
destruct (IHl x);auto.
right;intro.
destruct H;auto.
apply n;apply eq_nat_is_eq;auto.
Qed.

Fixpoint occ_block (E:list booleanEquation)(x:propVar) : Prop :=
match E with
| nil => False
| cons (bEqn _ f) E' => (uses f x) \/ (occ_block E' x)
end.

Fixpoint occ (E:list block)(x:propVar):Prop:=
match E with
| nil => False
| cons bl E' => occ_block bl x \/ occ E' x
end.

Fixpoint bnd_block_number (e:list block)(x:propVar):=
match e with
| nil => 0
| bl::e' => if (bnd_block_dec bl x) then 0 else S(bnd_block_number e' x)
end.

Lemma bnd_dec : forall e:list block, forall x:propVar, 
  {bnd e x} + {~bnd e x}.
Proof.
induction e;simpl in *;auto.

intros.
destruct (bnd_block_dec a x);auto.
destruct (IHe x);auto.
right;intro.
destruct H;auto.
Qed.

Lemma eq_dec_block : eq_dec (list booleanEquation).
Proof.
intro.
induction x;intros.
induction y.
left;auto.
right;intro.
inversion H.

induction y.
right;intro.
inversion H.

destruct (IHx y).
rewrite e in *.
destruct a;destruct a0.
destruct (eq_dec_nat p p1).
destruct (eq_dec_propForm p0 p2).
left;congruence.
right;intro.
apply n;congruence.
right;intro.
apply n;congruence.
right;intro.
apply n;congruence.
Qed.

(*well defined-ness*)

Definition w_d(E:list block) := forall x:propVar, bnd E x <-> bnd_cnt E x=1.

Lemma w_d_shorten : forall l:list block, forall b:block,
  w_d (b::l) -> w_d l.
Proof.
intros;intro.
destruct b as [b type_b exists_b].
destruct (H x).
simpl in *.
split;intro.
assert (bnd_cnt_block b x+bnd_cnt l x=1);auto.
assert (bnd_cnt l x>0).
  apply bnd_cnt_gt;auto.
  omega.
apply bnd_cnt_gt.
omega.
Qed.

Lemma w_d_bnd_list : forall E:list block, forall bl:_, w_d (bl::E) ->
  forall x:_, bnd E x -> ~bnd_block bl x.
Proof.
intros;intro;intros.
destruct (H x);simpl in *.
assert (bnd_block bl x \/ bnd E x);auto.
specialize (H2 H4).
assert (bnd_cnt_block bl x>0).
  apply bnd_cnt_block_gt;auto.
assert (bnd_cnt E x>0).
  apply bnd_cnt_gt;auto.
omega.
Qed.

Fixpoint rhs_block (l:list booleanEquation)(x:propVar):propForm :=
match l with 
| nil => var x
| cons (bEqn xc f) l' => if beq_nat xc x then f else rhs_block l' x
end.

Fixpoint rhs (l:list block)(x:propVar):propForm :=
match l with
| nil => bot
| bl::l' => if (bnd_block_dec bl x) 
            then rhs_block bl x
            else rhs l' x
end.

Lemma eq_bnd_block_number : forall e:list block, forall x y:propVar,
  bnd_block_number e x=bnd_block_number e y -> (bnd e x <-> bnd e y).
Proof.
induction e;simpl in *.
split;intro;auto.

split;intro.
destruct (bnd_block_dec a y);auto.
destruct (bnd_block_dec a x).
inversion H.
inversion H.
destruct (IHe _ _ H2);destruct H0;try contradiction;auto.
destruct (bnd_block_dec a x);auto.
destruct (bnd_block_dec a y).
inversion H.
inversion H.
destruct (IHe _ _ H2);destruct H0;try contradiction;auto.
Qed.

(*----------BES----------*)
Fixpoint alt_helper (E:_)(last_block_type:bool):=
match E with
| nil => True
| b::E1 => if eqb (blockType b) last_block_type then False else alt_helper E1 (blockType b)
end.

Definition alternating_list_block (E:_):=
match E with
| nil => True
| b::E2 => alt_helper E2 (blockType b)
end.

Record BES : Set := makeBES {
  bes:> list block;
  well_defined: w_d bes;
  alternating: alternating_list_block bes
}.

Definition first_block_type (E:_):bool:=
match E with
| nil => false
| b::E1 => blockType b
end.

Definition rank (E:BES)(x:propVar) : nat :=
if (bnd_dec E x) then
match (bes E) with
| nil => 0
| bl::_ => if first_block_type E
           then S(S(bnd_block_number E x))
           else S(bnd_block_number E x)
end
else 0.

Lemma eq_rank : forall e:BES, forall x y:propVar, rank e x=rank e y -> (bnd e x <-> bnd e y).
Proof.
unfold rank.
intros.
split;intros.
destruct (bnd_dec e x);try contradiction.
destruct (bnd_dec e y);auto.
destruct e as [e wd_e].
destruct e;simpl in *;try contradiction.
destruct b0 as [bl bt be].
destruct be;inversion H.
destruct (bnd_dec e y);try contradiction.
destruct (bnd_dec e x);auto.
destruct e as [e wd_e].
destruct e;simpl in *;try contradiction.
destruct b0 as [bl bt be].
destruct be;inversion H.
Qed.

Definition w_d_block (b:_):=forall x:_, bnd_block b x <-> bnd_cnt_block b x = 1.































