Add LoadPath projectPath.

Require Import Bool.
Require Import List.
Require Import Arith.
Require Import Omega.

Require Import propositional_formulas.
Require Import propositional_formulas_semantics.
Require Import BES.
Require Import relative_consequence.
Require Import consistent_consequence.

Lemma propForm_solution_dec_b : forall f:_, forall theta:_, {[[f]]theta}+{~[[f]]theta}.
Proof.
induction f;intros;simpl;auto.
right;intro;inversion H.

destruct (theta p);auto.
right;intro;inversion H.

destruct (IHf1 theta).
left;apply orb_true_iff;auto.
destruct (IHf2 theta).
left;apply orb_true_iff;auto.
right;intro;apply orb_true_iff in H;destruct H;auto.

destruct (IHf1 theta).
destruct (IHf2 theta).
left;apply andb_true_iff;split;auto.
right;intro;apply andb_true_iff in H;destruct H;auto.
right;intro;apply andb_true_iff in H;destruct H;auto.
Qed.

Fixpoint unfold_block (bl:list booleanEquation)(theta:environment):environment:=
match bl with
| nil => theta
| (bEqn x fx)::bl' => redefineEnvironment (unfold_block bl' theta) x (if (propForm_solution_dec_b fx theta) then true else false)
end.

Definition redefine_bound (bl:list booleanEquation)(result_unbound result_bound:environment):environment:=
  fun (x:propVar) => if bnd_block_dec bl x then result_bound x else result_unbound x.

Fixpoint sol_iterator (H:environment -> environment)(theta_b:environment)(i:nat) :=
match i with
| 0 => H theta_b
| S(k) => H (sol_iterator H theta_b k)
end.

Fixpoint genBES_solution (e:list block)(theta:environment) : environment :=
let F (e'':list block)(B'':block)(theta'' theta_b'':environment) : environment :=
  unfold_block B'' (genBES_solution e'' (redefine_bound B'' theta'' theta_b''))
in
match e with
| nil => theta
| cons bl e' => if (blockType bl)
                then genBES_solution e' (redefine_bound bl theta (sol_iterator (F e' bl theta) full_environment (length bl)))
                else genBES_solution e' (redefine_bound bl theta (sol_iterator (F e' bl theta) empty_environment (length bl)))
end.

Definition BES_solution (E:_)(Theta:environment) : environment :=
  genBES_solution E Theta.

Lemma neq_no_rewrite : forall theta:environment, forall x y:propVar, x<>y -> 
  forall b:bool, theta x <-> (redefineEnvironment theta y b) x.
Proof.
unfold redefineEnvironment.
split;intro.
assert (beq_nat y x=false).
  apply beq_nat_false_iff;auto.
rewrite H1;auto.
assert (beq_nat y x=false).
  apply beq_nat_false_iff;auto.
rewrite H1 in H0;auto.
Qed.

Lemma unbound_no_change_block_function : forall bl:list booleanEquation, forall x:propVar,
  ~bnd_block bl x -> forall theta:environment, theta x <-> unfold_block bl theta x.
Proof.
induction bl;simpl in *;auto.
split;intro;auto.

destruct a as [y fy].
split;intro.
assert (~bnd_block bl x).
  intro;apply H;auto.
apply neq_no_rewrite.
intro.
apply H;auto.
apply-> IHbl;auto.
apply neq_no_rewrite in H0.
apply IHbl;auto.
intro;apply H;auto.
Qed.

Lemma rank_bound_split: forall E:_, forall R:_, consistent_consequence E R ->
  forall x y:propVar, R (x,y) -> ((bnd E x /\ bnd E y) \/ (~bnd E x /\ ~bnd E y)).
Proof.
intros.
destruct H.
specialize (H _ H0);simpl in *.
destruct (eq_bnd_block_number _ _ _ H).
destruct (bnd_dec E x);auto.
right;split;intro;auto.
Qed.

Lemma alt_tl : forall l:list block, forall h:block, 
  alternating_list_block (h::l) -> alternating_list_block l.
Proof.
induction l;auto.
intros;simpl in *.
destruct (blockType h).
destruct (blockType a);try contradiction;auto.
destruct (blockType a);try contradiction;auto.
Qed.

Lemma w_d_w_d_block : forall bl:list booleanEquation,
  forall b:bool, forall exists_bl:_,
    (w_d ((makeBlock bl exists_bl b)::nil)) <-> w_d_block bl.
Proof.
destruct bl.
split;intros;destruct exists_bl;inversion i.

split;intros.
intro.
destruct b as [z fz].
destruct (H x).
simpl in *.
destruct (eq_nat_decide x z).
rewrite (eq_nat_eq _ _ e) in *.
rewrite<- beq_nat_refl in *.
assert (~bnd_block bl z).
  intro.
  assert (S (bnd_cnt_block bl z) > 1).
    apply gt_n_S.
    apply bnd_cnt_block_gt;auto.
  assert (S (bnd_cnt_block bl z) + 0 = 1).
    apply H0;auto.
  omega.
split;intro;auto.
assert (S (bnd_cnt_block bl z) + 0 = 1);auto.
omega.
assert (beq_nat x z=false).
  apply beq_nat_false_iff.
  intro.
  apply n;apply eq_nat_is_eq;auto.
split;intro.
destruct H3.
exfalso.
apply n;apply eq_nat_is_eq;auto.
rewrite H2 in *.
assert (bnd_cnt_block bl x + 0 = 1);auto.
omega.
rewrite H2 in *.
right.
apply bnd_cnt_block_gt;omega.

destruct b as [z fz].
intro;intros.
destruct (H x).
simpl in *.
destruct (eq_nat_decide x z).
rewrite (eq_nat_eq _ _ e) in *;auto.
rewrite<- beq_nat_refl in *.
split;intro;auto.
destruct H2;try contradiction.
assert (S (bnd_cnt_block bl z)=1);auto.
omega.
assert (beq_nat x z=false).
  apply beq_nat_false_iff;intro.
  apply n;apply eq_nat_is_eq;auto.
rewrite H2 in *.
split;intro.
destruct H3;try contradiction.
assert (bnd_cnt_block bl x=1);auto.
omega.
left.
apply H1.
omega.
Qed.

Lemma w_d_block_tl : forall bl:list booleanEquation, forall heqn:booleanEquation, 
  w_d_block (heqn::bl) -> w_d_block bl.
Proof.
destruct heqn as [z fz].
intros;intro.
destruct (H x).
simpl in *.
split;intro.
assert (beq_nat x z=false).
  apply beq_nat_false_iff.
  intro.
  rewrite H3 in *.
  rewrite<- beq_nat_refl in *.
  assert (S(bnd_cnt_block bl z)=1);auto.
  assert (S(bnd_cnt_block bl z)>1).
    apply gt_n_S.
    apply bnd_cnt_block_gt;auto.
  omega.
rewrite H3 in *;auto.
assert (beq_nat x z=false).
  apply beq_nat_false_iff.
  intro.
  rewrite H3 in *.
  rewrite<- beq_nat_refl in *.
  assert (S(bnd_cnt_block bl z)=1);auto.
  assert (S(bnd_cnt_block bl z)>1).
    apply gt_n_S.
    rewrite H2;auto.
  omega.
rewrite H3 in *.
destruct (H1 H2);auto.
exfalso.
apply beq_nat_false_iff in H3;auto.
Qed.

Lemma rhs_unfold_block : forall bl:block, w_d (bl::nil) ->
      forall x:propVar, bnd_block bl x -> 
        forall theta:environment, 
          (unfold_block bl theta) x
        <->
          propForm_solution (rhs (bl::nil) x) theta.
Proof.
destruct bl as [bl exists_bl type_bl].
induction bl;simpl in *;auto.
intros;contradiction.

destruct a as [z fz].
intros.
simpl in *.
destruct (bnd_block_dec (bEqn z fz :: bl) x);try contradiction.
destruct (eq_nat_decide x z).
rewrite (eq_nat_eq _ _ e) in *;auto.
unfold redefineEnvironment.
rewrite<- beq_nat_refl.
destruct (propForm_solution_dec_b fz);split;auto.
intro;inversion H1.
unfold redefineEnvironment.
assert (beq_nat z x=false).
  apply beq_nat_false_iff.
  intro.
  apply n;apply eq_nat_is_eq;auto.
repeat rewrite H1.
destruct H0.
exfalso.
apply n;apply eq_nat_is_eq;auto.
assert (exists beqn : booleanEquation, In beqn bl).
  clear - b n.
  destruct bl;simpl in *.
  exfalso.
  destruct b;auto.
  apply n;apply eq_nat_is_eq;auto.
  exists b0;auto.
assert (type_bl=type_bl /\ True);auto.
assert (w_d ((makeBlock bl H2 type_bl)::nil)).
  apply w_d_w_d_block.
  apply (w_d_block_tl bl (bEqn z fz)).
  apply (w_d_w_d_block (bEqn z fz::bl) type_bl exists_bl);auto.
assert True;auto.
destruct (IHbl H2 H4 x H0 theta).
destruct (bnd_block_dec bl x);try contradiction.
split;intro;auto.
Qed.

Lemma unfold_block_maintains_consistency : forall bl:block,
    forall w_d_bl: w_d (bl::nil),
      forall R:_, 
        consistent_consequence (bl::nil) R ->
forall theta:environment, consistent_environment theta R -> 
  consistent_environment (unfold_block bl theta) R.
Proof.
destruct bl as [bl exists_bl type_bl].
destruct bl;simpl;auto.

destruct b as [z fz].
intros.
intro;intros;simpl in *.
destruct (rank_bound_split _ _ H _ _ H1).
destruct H3.
destruct H3;try contradiction.
destruct H4;try contradiction.

apply (rhs_unfold_block (makeBlock (bEqn z fz::bl) exists_bl type_bl) w_d_bl y H4 theta);simpl in *.
destruct (bnd_block_dec (bEqn z fz :: bl) y);try contradiction.

destruct (rhs_unfold_block (makeBlock (bEqn z fz::bl) exists_bl type_bl) w_d_bl x H3 theta);simpl in *.
destruct (bnd_block_dec (bEqn z fz :: bl) x);try contradiction.
specialize (H5 H2).
destruct H.
specialize (H7 _ H1 _ H0);simpl in H7.
destruct (bnd_block_dec (bEqn z fz :: bl) y);try contradiction.
destruct (bnd_block_dec (bEqn z fz :: bl) x);try contradiction;auto.
destruct H3.
unfold redefineEnvironment in *.
assert (beq_nat z x=false).
  apply beq_nat_false_iff;intro.
  apply H3;simpl;auto.
assert (beq_nat z y=false).
  apply beq_nat_false_iff;intro.
  apply H4;simpl;auto.
rewrite H6.
rewrite H5 in H2.
assert (~bnd_block bl x).
  intro;apply H3;simpl;auto.
assert (~bnd_block bl y).
  intro;apply H4;simpl;auto.
apply (unbound_no_change_block_function _ _ H8 theta).
apply (unbound_no_change_block_function _ _ H7) in H2.
apply (H0 x);auto.
Qed.

Lemma sol_iterator_maintains_consistency : forall f:environment -> environment, forall R:_, 
(forall theta:environment, consistent_environment theta R -> (consistent_environment (f theta) R)) ->
  forall n:nat, forall theta:environment, consistent_environment theta R -> 
    consistent_environment (sol_iterator f theta n) R.
Proof.
induction n;simpl in *;auto.
Qed.

Lemma w_d_hd : forall bl:list block, forall bh:block, 
  w_d (bh::bl) -> w_d_block bh.
Proof.
intros;intro.
destruct bh as [bh type_bh exists_bh].
destruct (H x).
simpl in *.
split;intro.
assert (bnd_cnt_block bh x+bnd_cnt bl x=1);auto.
assert (bnd_cnt_block bh x>0).
  apply bnd_cnt_block_gt;auto.
  omega.
apply bnd_cnt_block_gt.
omega.
Qed.

Lemma cc_tl : forall e:list block, forall bl:block, 
forall w_d_bl_e: w_d (bl::e),
  forall R:_, consistent_consequence (bl::e) R ->
      consistent_consequence e R.
Proof.
intros.
split.
intro;intros.
destruct H.
specialize (H _ H0).
simpl in *.
destruct (bnd_block_dec bl (fst XY)).
destruct (bnd_block_dec bl (snd XY)).
assert (~bnd e (fst XY)).
  clear - b w_d_bl_e.
  intro.
  destruct (w_d_bl_e (fst XY)).
  assert (bnd (bl :: e) (fst XY));simpl;auto.
  specialize (H0 H2).
  destruct (bnd_cnt_gt (bl::nil) (fst XY)).
  destruct (bnd_cnt_gt e (fst XY)).
  assert (bnd (bl :: nil) (fst XY));simpl;auto.
  specialize (H3 H7).
  specialize (H5 H).
  simpl in *;clear - H0 H3 H5.
  omega.
assert (~bnd e (snd XY)).
  clear - b0 w_d_bl_e.
  intro.
  destruct (w_d_bl_e (snd XY)).
  assert (bnd (bl :: e) (snd XY));simpl;auto.
  specialize (H0 H2).
  destruct (bnd_cnt_gt (bl::nil) (snd XY)).
  destruct (bnd_cnt_gt e (snd XY)).
  assert (bnd (bl :: nil) (snd XY));simpl;auto.
  specialize (H3 H7).
  specialize (H5 H).
  simpl in *;clear - H0 H3 H5.
  omega.
clear - H2 H3.
induction e;simpl in *;auto.
rewrite IHe;auto.
destruct (bnd_block_dec a (fst XY));destruct (bnd_block_dec a (snd XY));auto.
exfalso;apply H2;auto.
exfalso;apply H3;auto.
inversion H.
destruct (bnd_block_dec bl (snd XY));inversion H;auto.

intro;intros.
intro;intros.
simpl in *.
destruct H.
specialize (H3 _ H0 _ H1).
simpl in *.
destruct (bnd_block_dec bl (fst XY)).
destruct (bnd_block_dec bl (snd XY)).
assert (~bnd e (fst XY)).
  clear - b w_d_bl_e.
  intro.
  destruct (w_d_bl_e (fst XY)).
  assert (bnd (bl :: e) (fst XY));simpl;auto.
  specialize (H0 H2).
  destruct (bnd_cnt_gt (bl::nil) (fst XY)).
  destruct (bnd_cnt_gt e (fst XY)).
  assert (bnd (bl :: nil) (fst XY));simpl;auto.
  specialize (H3 H7).
  specialize (H5 H).
  simpl in *;clear - H0 H3 H5.
  omega.
assert (~bnd e (snd XY)).
  clear - b0 w_d_bl_e.
  intro.
  destruct (w_d_bl_e (snd XY)).
  assert (bnd (bl :: e) (snd XY));simpl;auto.
  specialize (H0 H2).
  destruct (bnd_cnt_gt (bl::nil) (snd XY)).
  destruct (bnd_cnt_gt e (snd XY)).
  assert (bnd (bl :: nil) (snd XY));simpl;auto.
  specialize (H3 H7).
  specialize (H5 H).
  simpl in *;clear - H0 H3 H5.
  omega.
clear - H2 H4 H5.
induction e;simpl in *;auto.
assert (~ bnd e (fst XY)).
  intro;apply H4;auto.
assert (~ bnd e (snd XY)).
  intro;apply H5;auto.
destruct (bnd_block_dec a (snd XY));auto.
apply IHe;auto.
destruct (bnd_block_dec a (fst XY));auto.
specialize (H _ H0).
exfalso.
clear - H b n.
simpl in H.
destruct (bnd_block_dec bl (fst XY));try contradiction.
destruct (bnd_block_dec bl (snd XY));try contradiction.
inversion H.
destruct (bnd_block_dec bl (snd XY));auto.
specialize (H _ H0).
exfalso.
clear - H b n.
simpl in H.
destruct (bnd_block_dec bl (fst XY));try contradiction.
destruct (bnd_block_dec bl (snd XY));try contradiction.
inversion H.
Qed.

Lemma cc_hd : forall e:_, forall bl:_,
forall w_d_bl_e: w_d (bl::e),
  forall R:_, consistent_consequence (bl::e) R -> consistent_consequence (bl::nil) R.
Proof.
intros.
destruct H.
split.
intro;intros.
specialize (H _ H1).
simpl in *.
destruct (bnd_block_dec bl (fst XY)).
destruct (bnd_block_dec bl (snd XY));auto.
inversion H.
destruct (bnd_block_dec bl (snd XY));auto.
inversion H.

intro;intros.
intro;intros.
specialize (H _ H1).
specialize (H0 _ H1 _ H2);simpl in *.
destruct (bnd_block_dec bl (fst XY)).
destruct (bnd_block_dec bl (snd XY));auto.
inversion H.
destruct (bnd_block_dec bl (snd XY));auto.
inversion H.
Qed.

Lemma redef_bnd_consistent : forall bl:_, forall R:_, consistent_consequence (bl::nil) R ->
  forall theta_bound theta_unbound:_, consistent_environment theta_bound R ->
  consistent_environment theta_unbound R ->
    consistent_environment (redefine_bound bl theta_bound theta_unbound) R.
Proof.
intros;intro;intros.
unfold redefine_bound in *.
destruct H.
specialize (H _ H2).
simpl in *.
destruct (bnd_block_dec bl x).
destruct (bnd_block_dec bl y).
apply (H1 x);auto.
inversion H.
destruct (bnd_block_dec bl y).
inversion H.
apply (H0 x);auto.
Qed.

Lemma BES_solution_maintains_consistency : forall E:_, w_d E -> forall R:_,
  consistent_consequence E R ->
    forall theta:environment, consistent_environment theta R ->
      consistent_environment (BES_solution E theta) R.
Proof.
induction E as [|a e].
simpl in *;auto.

simpl;intros.
assert (w_d e).
  apply (w_d_shorten e a);auto.
assert (consistent_consequence e R).
  apply (cc_tl e a H);auto.
assert (consistent_environment (genBES_solution e theta) R).
  apply (IHe H2 R H3 theta);auto.
assert (w_d (a::nil)).
  destruct a as [al].
  apply w_d_w_d_block.
  apply (w_d_hd _ _ H).
assert (consistent_consequence (a::nil) R).
  apply (cc_hd e a);auto.
destruct a as [bl bl_exists bl_type].
simpl.
destruct bl_type.

(*nu*)
cut (consistent_environment (redefine_bound bl theta
        (sol_iterator
           (fun theta_b'' : environment =>
            unfold_block bl
              (genBES_solution e (redefine_bound bl theta theta_b'')))
           full_environment (length bl))) R).
intro.
apply (IHe H2 R H3);auto.
apply (redef_bnd_consistent _ _ H6 _ _ H1).
apply sol_iterator_maintains_consistency.
intros.
cut (consistent_environment (genBES_solution e (redefine_bound bl theta theta0)) R).
intro.
apply (unfold_block_maintains_consistency _ H5 _ H6);auto.
apply (IHe H2 R H3);auto.
apply (redef_bnd_consistent _ _ H6);auto.
apply full_environment_consistent.

(*mu*)
cut (consistent_environment (redefine_bound bl theta
        (sol_iterator
           (fun theta_b'' : environment =>
            unfold_block bl
              (genBES_solution e (redefine_bound bl theta theta_b'')))
           empty_environment (length bl))) R).
intro.
apply (IHe H2 R H3);auto.
apply (redef_bnd_consistent _ _ H6 _ _ H1).
apply sol_iterator_maintains_consistency.
intros.
cut (consistent_environment (genBES_solution e (redefine_bound bl theta theta0)) R).
intro.
apply (unfold_block_maintains_consistency _ H5 _ H6);auto.
apply (IHe H2 R H3);auto.
apply (redef_bnd_consistent _ _ H6);auto.
apply empty_environment_consistent.
Qed.

Lemma cc_BES_semantics : forall E:BES, forall R:_, consistent_consequence E R ->
  forall theta:environment, consistent_environment theta R ->
    consistent_environment (BES_solution E theta) R.
Proof.
destruct E as [E wd_E alt_E].
apply BES_solution_maintains_consistency;auto.
Qed.


Definition lteq_environment(theta1 theta2:environment) :=
  forall x:_, theta1 x=true -> theta2 x=true.

Lemma unfold_block_monotone : forall b:_, w_d_block b -> forall theta1 theta2:_,
  lteq_environment theta1 theta2 -> lteq_environment (unfold_block b theta1) (unfold_block b theta2).
Proof.
induction b;simpl in *;auto;intros.
destruct a as [y fy].
specialize (IHb (w_d_block_tl _ _ H)).
assert (propForm_solution fy theta1 -> propForm_solution fy theta2).
  clear - H0.
  induction fy;simpl in *;auto.
  intro.
  apply orb_true_iff in H;destruct H;apply orb_true_iff.
  left;apply IHfy1;auto.
  right;apply IHfy2;auto.
  intro;apply andb_true_iff in H;destruct H;apply andb_true_iff;split.
  apply IHfy1;auto.
  apply IHfy2;auto.
intro.
unfold redefineEnvironment.
destruct beq_nat.
destruct propForm_solution_dec_b.
destruct propForm_solution_dec_b;auto.
intro.
inversion H2.
apply (IHb _ _ H0).
Qed.

