Add LoadPath projectPath.

Require Import relation_definitions.
Require Import propositional_formulas.
Require Import propositional_formulas_semantics.
Require Import relative_consequence.
Require Import relations_on_BES.
Require Import Ensembles.

(*----------consistent consequences----------*)

Definition consistent_consequence (E:_)(R:relation propVar) : Prop :=
  rel_eq_rank E R /\ rhs_cons E R.

Definition cc_max (E:_):relation propVar :=
  fun (XY:_) =>
    exists R:relation propVar, consistent_consequence E R /\ relates_bound E R /\ R XY.

Notation "E |- x <_ y" := (cc_max E (x,y))(at level 50).

Lemma cc_max_consistent_environment_has_cc: forall E:_, forall theta:environment, 
consistent_environment theta (cc_max E) ->
  forall x y:propVar, (E |- x <_ y) ->
    exists R:relation propVar,
      consistent_consequence E R /\ R (x,y) /\ consistent_environment theta R.
Proof.
intros.
destruct H0 as [R H0].
destruct H0.
destruct H1.
exists R.
split;auto.
split;auto.
intro;intros.
apply (H x0);auto.
exists R;auto.
Qed.

Lemma cc_max_rel_eq_rank : forall E:_, rel_eq_rank E (cc_max E).
Proof.
intros;intro;intros.
destruct H as [R H0].
destruct H0.
destruct H0.
destruct H;auto.
Qed.

Lemma cc_max_rhs_cons : forall E:_, rhs_cons E (cc_max E).
Proof.
intros;intro;intros;intro;intros.
destruct XY as (x,y).
destruct (cc_max_consistent_environment_has_cc E theta H0 _ _ H) as [R H2].
destruct H2.
destruct H3.
simpl in *.
destruct H2.
apply (H5 (x,y));auto.
Qed.

Lemma cc_max_is_cc : forall E:_, consistent_consequence E (cc_max E).
Proof.
split.
apply cc_max_rel_eq_rank.
apply cc_max_rhs_cons.
Qed.

Lemma union_maintains_cc : forall E:_, forall R1 R2:_, (consistent_consequence E R1) ->
  (consistent_consequence E R2) -> (consistent_consequence E (Union _ R1 R2)).
Proof.
intros.
destruct H.
destruct H0.
split;intro;intros.
destruct H3;auto.
intro;intros.
destruct (consistent_environment_rel_union _ _ _ H4);auto.
destruct H3.
apply H1;auto.
apply H2;auto.
Qed.

Lemma cc_max_is_max : forall E:_, forall R:_, (consistent_consequence E R) ->
  relates_bound E R -> Included _ R (cc_max E).
Proof.
intros;intro;intros.
exists R;auto.
Qed.




































