Add LoadPath projectPath.

Require Import Finite_sets.
Require Import Bool.
Require Import Ensembles.
Require Import relation_definitions.
Require Import relation_lemmas.
Require Import propositional_formulas.
Require Import propositional_formulas_semantics.
Require Import consequence_propositional_formula.
Require Import propositional_formula_DNF.
Require Import transitive_closure_definitions.
Require Import transitive_closure_minimal_environment.
Require Import relative_consequence.
Require Import relative_consequence_minimal_environment.
Require Import relation_induction_on_unrelated.
Require Import BES.
Require Import proofSystem_consistent_consequence.
Require Import proofSystem_complete_logical_consequence.
Require Import relations_on_BES.
Require Import consistent_consequence.
Require Import relative_consistent_consequence.
Require Import consistent_consequence_excluded_middle.

(*----------completeness----------*)

Lemma prv_tree_reachable : forall E:_, forall R G:relation propVar,
  forall XY:_, reachable _ R XY -> (forall x y : propVar, R (x, y) -> prv_tree E (G |- var x --> var y)) ->
    prv_tree E (G |- var (fst XY) --> var (snd XY)).
Proof.
intros.
induction H;simpl in *;auto.

apply (TRA _ _ (var y));auto.
Qed.

Lemma complete_fom_conj_rel_cons : forall E:_, forall R G:relation propVar,
Finite _ R ->
  forall f g:propForm, conj_propForm f -> DNF_propForm g ->
    rel_cons R f g ->
      (forall x y:propVar, R (x,y) -> prv_tree E (stmt G (var x) (var y))) ->
        prv_tree E (stmt G f g).
Proof.
intros.
induction H1 using DNF_ind;simpl in *.
apply complete_cons.
intro;auto.

apply complete_to_bot.
intro;intros.
apply (full_false_eqv_bot f);auto.
intro.
assert false.
  apply (H2 _ (full_environment_consistent _));auto.
inversion H5.

induction H0 using conj_ind.
  assert false.
    apply (H2 _ (empty_environment_consistent _));auto.
  inversion H0.

  apply complete_from_bot.

  destruct (exists_list _ _ H) as [l H4].
  specialize (H2 (reachable_from _ eq_dec_nat _ _ H4 x0) (reachable_from_consistent _ _ H4 x0) (reachable_from_self _ _ H4 x0)).
  unfold reachable_from in H2;simpl in H2.
  destruct (reachable_dec_b _ eq_dec_nat _ _ H4 x0 x).
  apply (prv_tree_reachable _ _ _ (x0,x) r);auto.
  destruct (eq_dec_nat x0 x).
  rewrite e;apply REF.
  inversion H2.

  destruct (propForm_solution_choice (f /\p g) full_environment).
  assert (~propForm_eqv (f /\p g) bot).
    intro.
    assert false.
      apply (H1 full_environment);auto.
    inversion H4.
  destruct (exists_list _ _ H) as [l H4].
  specialize (H2 (min_env_for_propForm _ _ H4 (f /\p g)) (min_env_for_propForm_consistent _ _ H4 _) (min_env_for_propForm_self _ _ H4 _ H1)).
  simpl in H2.
  unfold environment_union in H2;apply orb_true_iff in H2;destruct H2.
  apply aGROW_L.
  apply IHconj_propForm1.
  apply (min_env_conj _ _ _ H4 _ H0_);auto.
  intro.
  apply H1.
  split;intros.
  apply H5.
  apply andb_true_iff in H6;destruct H6;auto.
  inversion H6.
  apply aGROW_R.
  apply IHconj_propForm2.
  apply (min_env_conj _ _ _ H4 _ H0_0);auto.
  intro.
  apply H1.
  split;intros.
  apply H5.
  apply andb_true_iff in H6;destruct H6;auto.
  inversion H6.

  apply (TRA _ _ bot).
  apply complete_cons.
  intro;intros.
  apply (full_false_eqv_bot _ H0);auto.
  apply complete_from_bot.

apply cSPLIT.
apply IHDNF_propForm.
intro;intros.
specialize (H2 _ H5 H6).
apply andb_true_iff in H2;destruct H2;auto.
apply IHDNF_propForm0.
intro;intros.
specialize (H2 _ H5 H6).
apply andb_true_iff in H2;destruct H2;auto.

destruct (propForm_solution_choice f full_environment).
assert (~propForm_eqv f bot).
  intro.
  assert false.
    apply (H4 full_environment);auto.
  inversion H5.
destruct (exists_list _ _ H) as [l H5].
specialize (H2 (min_env_for_propForm _ _ H5 f) (min_env_for_propForm_consistent _ _ H5 _) (min_env_for_propForm_self _ _ H5 _ H4)).
simpl in H2;apply orb_true_iff in H2;destruct H2.
apply cGROW_L.
apply IHDNF_propForm1.
apply (min_env_conj _ _ _ H5);auto.
apply cGROW_R.
apply IHDNF_propForm2.
apply (min_env_conj _ _ _ H5);auto.
apply (TRA _ _ bot).
apply complete_cons.
intro;intros.
apply (full_false_eqv_bot _ H1);auto.
apply complete_from_bot.
Qed.

Lemma derivable_set_derivable_rel_cons_DNF : forall E:_, forall R G:relation propVar,
  Finite _ R ->
    forall f g:propForm, DNF_propForm f -> DNF_propForm g ->
      rel_cons R f g ->
        (forall x y:_, R (x,y) -> prv_tree E (stmt G (var x) (var y))) ->
          prv_tree E (stmt G f g).
Proof.
intros.
induction H0 using DNF_ind.
apply (complete_fom_conj_rel_cons _ _ _ H);simpl;auto.
apply (complete_fom_conj_rel_cons _ _ _ H);simpl;auto.
apply (complete_fom_conj_rel_cons _ _ _ H);simpl;auto.
apply (complete_fom_conj_rel_cons _ _ _ H);simpl;auto.

assert (rel_cons R f g).
  intro;intros.
  apply H2;auto.
  apply orb_true_iff;auto.
assert (rel_cons R g0 g).
  intro;intros.
  apply H2;auto.
  apply orb_true_iff;auto.

apply aSPLIT;auto.
Qed.

Lemma derivable_set_derivable_rel_cons : forall E:_, forall R G:relation propVar,
  Finite _ R ->
    forall f g:propForm, rel_cons R f g ->
        (forall x y:propVar, R (x,y) -> prv_tree E (stmt G (var x) (var y))) ->
          prv_tree E (stmt G f g).
Proof.
intros.
apply (TRA _ _ (toDNF f)).
apply complete_cons.
intro;intros;apply (toDNF_eqv f);auto.
apply (TRA _ _ (toDNF g)).
apply (derivable_set_derivable_rel_cons_DNF _ _ _ H);auto.
apply DNF_toDNF.
apply DNF_toDNF.
intro;intros.
apply (toDNF_eqv g).
apply H0;auto.
apply (toDNF_eqv f);auto.
apply complete_cons.
intro;intros.
apply (toDNF_eqv g);auto.
Qed.

Lemma complete_propVar : forall n:nat, forall E:_, w_d E -> forall G:relation propVar,
  rev_subset_card _ (rel_all_bnd_eq_rank E) G n ->
    forall XY:_, rel_cc_max E G XY -> prv_tree E (stmt G (var (fst XY)) (var (snd XY))).
Proof.
induction n using nat_rec;intros.
assert (Same_set _ (rel_all_bnd_eq_rank E) G).
  apply (rev_subset_card0 _ _ _ _ H0);auto.
destruct H1 as [R H1].
rewrite<- (Extensionality_Ensembles _ _ _ H2) in *.
apply CNT.
destruct H1;destruct H3;destruct H4;destruct H5;auto.
split;destruct x.
apply H1;auto.
apply H3;auto.

assert (Finite _ G).
  assert (Finite _ (rel_all_bnd_eq_rank E)).
    apply rel_all_eq_rank_bnd_finite;auto.
  apply (rev_subset_card_finite _ _ _ _ eq_dec_nat H2 H0).
assert (G XY \/ ~G XY).
  apply Finite_dec;auto.
  apply (eq_dec_pairs _ eq_dec_nat).
destruct H3.
apply CNT.
destruct XY;simpl;auto.
apply CC.
destruct H1.
destruct H1.
destruct H4.
destruct H5.
destruct H6;try contradiction.
destruct (H4 _ H6).
specialize (H1 _ H6).
split;auto.
assert (rev_subset_card _ (rel_all_bnd_eq_rank E) (Add _ G XY) n).
  assert (Same_set _ G (Subtract _ (Add _ G XY) XY)).
    split;intro;intros.
    split;auto.
    left;auto.
    intro.
    assert (XY = x).
      elim H5;auto.
    rewrite H6 in *;apply H3;auto.
    destruct H4;auto.
    destruct H4;auto.
    contradiction.
  rewrite (Extensionality_Ensembles _ _ _ H4) in H0.
  assert (rel_all_bnd_eq_rank E XY).
    destruct H1.
    destruct H1.
    destruct H5.
    destruct H6.
    destruct H7;try contradiction.
    split.
    apply H1;auto.
    apply H5;auto.
  assert (~(Subtract (nat * nat) (Add (nat * nat) G XY) XY) XY).
    intro.
    destruct H6.
    apply H7;apply In_singleton.
  specialize (rev_subset_card_minus _ (eq_dec_pairs _ eq_dec_nat) _ _  (rel_all_eq_rank_bnd_finite _ H)  _ H0 _ H5 H6);intro.
  simpl in H7.
  rewrite (Extensionality_Ensembles _ (Add _ G XY) (Add _ (Subtract _ (Add _ G XY) XY) XY)).
  assert (n=n-0).
    destruct n;auto.
  rewrite H8;auto.
  split;intro;intros.
  destruct (eq_dec_pairs _ eq_dec_nat XY x).
  rewrite e in *.
  right;apply In_singleton.
  left;split;auto.
  intro;apply n0;elim H9;auto.
  destruct (eq_dec_pairs _ eq_dec_nat XY x).
  rewrite e in *.
  right;apply In_singleton.
  left.
  destruct H8.
  destruct H8.
  destruct H8;auto.
  contradiction.
  exfalso;apply n0;elim H8;auto.
destruct (rel_cc_exists_finite _ H _ (rev_subset_card_rev_subset _ _ _ _ H0) _ H1) as [R H5].
apply (derivable_set_derivable_rel_cons _ (Union _ R G)).
destruct H5.
apply finite_union;auto.
apply (eq_dec_pairs _ eq_dec_nat).
destruct H5.
destruct H6.
destruct H7.
destruct H7.
intro;intros;apply H9;auto.
destruct H6;auto;contradiction.
intros.
assert (rel_cc_max E (Add _ G XY) (x,y)).
  exists R.
  destruct H5;destruct H7;destruct H8.
  destruct H8.
  split;auto.
  split;auto.
  split.
  intro;intros.
  intro;intros.
  apply H10;auto.
  intro;intros.
  apply (H12 x0);auto.
  destruct H14.
  apply Union_introl;auto.
  apply Union_intror;left;auto.
  destruct H6.
  apply Union_introl;auto.
  apply Union_intror;left;auto.
specialize (IHn _ H _ H4 _ H7);simpl in *.
rewrite (Extensionality_Ensembles _ (Add _ G XY) (Union _ G (Singleton _ XY))) in IHn.
destruct XY;simpl in *;auto.
split;intro;intros.
destruct H8.
apply Union_introl;auto.
apply Union_intror;auto.
destruct H8.
left;auto.
right;auto.
Qed.

Lemma complete : forall E:_, w_d E ->
    forall XY:_, cc_max E XY -> prv_tree E (stmt (Empty_set _) (var (fst XY)) (var (snd XY))).
Proof.
intros.
rewrite (eq_cc_max_rel_cc_max E) in H0.
destruct (exists_rev_subset _ _ _ (empty_rev_subset _ (eq_dec_pairs _ eq_dec_nat) _ (rel_all_eq_rank_bnd_finite _ H))) as [n H2].
apply (complete_propVar n);auto.
Qed.

