Add LoadPath projectPath.

Require Import relation_definitions.
Require Import relation_lemmas.
Require Import propositional_formulas.
Require Import BES.
Require Import Ensembles.
Require Import Arith.

(*----------prv_tree----------*)

Inductive statement := stmt : relation propVar -> propForm -> propForm -> statement.

Notation "G |- f --> g" := (stmt G f g)(at level 50).

Inductive prv_tree (E:_) : statement -> Prop :=
| AS1 (a b c:propForm)(G:relation propVar):
  (prv_tree E (stmt G (andp a (andp b c)) (andp (andp a b) c)))
| AS2 (a b c:propForm)(G:relation propVar):
  (prv_tree E (stmt G (andp (andp a b) c) (andp a (andp b c))))
| AS3 (a b c:propForm)(G:relation propVar):
  (prv_tree E (stmt G (orp a (orp b c)) (orp (orp a b) c)))
| AS4 (a b c:propForm)(G:relation propVar):
  (prv_tree E (stmt G (orp (orp a b) c) (orp a (orp b c))))

| COM1 (a b:propForm)(G:relation propVar):
  (prv_tree E (stmt G (andp a b) (andp b a)))
| COM2 (a b:propForm)(G:relation propVar):
  (prv_tree E (stmt G (orp a b) (orp b a)))

| DS1 (a b c:propForm)(G:relation propVar):
  (prv_tree E (stmt G (orp a (andp b c)) (andp (orp a b) (orp a c))))
| DS2 (a b c:propForm)(G:relation propVar):
  (prv_tree E (stmt G (andp (orp a b) (orp a c)) (orp a (andp b c))))
| DS3 (a b c:propForm)(G:relation propVar):
  (prv_tree E (stmt G (andp a (orp b c)) (orp (andp a b) (andp a c))))
| DS4 (a b c:propForm)(G:relation propVar):
  (prv_tree E (stmt G (orp (andp a b) (andp a c)) (andp a (orp b c))))

| AB1 (a b:propForm)(G:relation propVar):
  (prv_tree E (stmt G (orp a (andp a b)) a))
| AB2 (a b:propForm)(G:relation propVar):
  (prv_tree E (stmt G a (orp a (andp a b))))

| ID1 (a:propForm)(G:relation propVar):
  (prv_tree E (stmt G a (andp a a)))
| ID2 (a:propForm)(G:relation propVar):
  (prv_tree E (stmt G (orp a a) a))

| SUP (a b:propForm)(G:relation propVar):
  (prv_tree E (stmt G a (orp a b)))
| INF (a b:propForm)(G:relation propVar):
  (prv_tree E (stmt G (andp a b) a))

| TOP (a:propForm)(G:relation propVar):
  (prv_tree E (stmt G a (andp a top)))
| BOT (a:propForm)(G:relation propVar):
  (prv_tree E (stmt G (orp a bot) a))

| CTX (a b c:propForm)(x:propVar)(G:relation propVar):
  (prv_tree E (stmt G a b) -> prv_tree E (stmt G (replace c x a) (replace c x b)))
| TRA (a b c:propForm)(G:relation propVar):
  (prv_tree E (stmt G a b) -> prv_tree E (stmt G b c) -> prv_tree E (stmt G a c))
| REF (a:propForm)(G:relation propVar):
  (prv_tree E (stmt G a a))

| CC (x y:propVar)(G:relation propVar):
  (((bnd E x) /\ (bnd E y) /\ (bnd_block_number E x=bnd_block_number E y)) -> 
    (prv_tree E (stmt (Union _ G (Singleton _ (x,y))) (rhs E x)(rhs E y))) -> 
      (prv_tree E (stmt G (var x) (var y))))
| CNT (x y:propVar)(G:relation propVar):
  ((G (x,y)) -> (prv_tree E (stmt G (var x) (var y))))
.

(*derived rules*)

Lemma cSPLIT : forall E:_, forall G:_, forall f g1 g2:propForm, 
  prv_tree E (stmt G f g1) -> prv_tree E (stmt G f g2) -> prv_tree E (stmt G f (g1 /\p g2)).
Proof.
intros.
apply (TRA _ _ (f /\p g2)).
apply (TRA _ _ (f /\p f)).
apply ID1.
destruct (exists_unused f) as [fx Hx].
specialize (CTX _ f g2 (f /\p var fx) fx _ H0);auto;intro.
simpl in H1.
rewrite <- beq_nat_refl in H1.
rewrite unused_no_rewrite in H1;auto.
rewrite unused_no_rewrite in H1;auto.
destruct (exists_unused g2) as [gx Hx].
specialize (CTX _ f g1 (var gx /\p g2) gx _ H);intro.
simpl in H1.
rewrite <- beq_nat_refl in H1.
rewrite unused_no_rewrite in H1;auto.
rewrite unused_no_rewrite in H1;auto.
Qed.

Lemma cGROW_L : forall E:_, forall G:_, forall f g1 g2:propForm, 
  prv_tree E (stmt G f g1) -> prv_tree E (stmt G f (g1 \/p g2)).
Proof.
intros.
apply (TRA _ _ (f \/p g2)).
apply SUP.
destruct (exists_unused g2) as [gx Hx].
specialize (CTX _ f g1 (var gx \/p g2) gx _ H);intro.
simpl in H0.
rewrite <- beq_nat_refl in H0.
rewrite unused_no_rewrite in H0;auto.
rewrite unused_no_rewrite in H0;auto.
Qed.

Lemma cGROW_R : forall E:_, forall G:relation propVar, forall f g1 g2:propForm, 
  prv_tree E (stmt G f g2) -> prv_tree E (stmt G f (g1 \/p g2)).
Proof.
intros.
apply (TRA _ _ (g2 \/p g1)).
apply cGROW_L;auto.
apply COM2.
Qed.

Lemma aSPLIT : forall E:_, forall G:relation propVar, forall f1 f2 g:propForm, 
  prv_tree E (stmt G f1 g) -> prv_tree E (stmt G f2 g) -> prv_tree E (stmt G (f1 \/p f2) g).
Proof.
intros.
apply (TRA _ _ (f1 \/p g)).
destruct (exists_unused f1) as [fx Hx].
specialize (CTX _ f2 g (f1 \/p var fx) fx _ H0);intro.
simpl in H1.
rewrite <- beq_nat_refl in H1.
rewrite unused_no_rewrite in H1;auto.
rewrite unused_no_rewrite in H1;auto.
apply (TRA _ _ (g \/p g)).
destruct (exists_unused g) as [gx Hx].
specialize (CTX _ f1 g (var gx \/p g) gx _ H);intro.
simpl in H1.
rewrite <- beq_nat_refl in H1.
rewrite unused_no_rewrite in H1;auto.
rewrite unused_no_rewrite in H1;auto.
apply ID2.
Qed.

Lemma aGROW_L : forall E:_, forall G:relation propVar, forall f1 f2 g:propForm, 
  prv_tree E (stmt G f1 g) -> prv_tree E (stmt G (f1 /\p f2) g).
Proof.
intros.
apply (TRA _ _ (g /\p f2)).
destruct (exists_unused f2) as [fx Hx].
specialize (CTX _ f1 g (var fx /\p f2) fx _ H);intro.
simpl in H0.
rewrite <- beq_nat_refl in H0.
rewrite unused_no_rewrite in H0;auto.
rewrite unused_no_rewrite in H0;auto.
apply INF.
Qed.

Lemma aGROW_R : forall E:_, forall G:relation propVar, forall f1 f2 g:propForm, 
  prv_tree E (stmt G f2 g) -> prv_tree E (stmt G (f1 /\p f2) g).
Proof.
intros.
apply (TRA _ _ (f2 /\p f1)).
apply COM1.
apply aGROW_L;auto.
Qed.

