Add LoadPath projectPath.

Require Import Ensembles.
Require Import Bool.
Require Import Finite_sets.

Require Import relation_definitions.
Require Import relation_lemmas.
Require Import propositional_formulas.
Require Import transitive_closure_definitions.
Require Import transitive_closure_minimal_environment.
Require Import relative_consequence.
Require Import relative_consequence_minimal_environment.
Require Import BES.
Require Import proofSystem_consistent_consequence.
Require Import consistent_consequence.
Require Import relative_consistent_consequence.

Lemma AS1_sound : forall E:_, forall a b c:_, forall G:_, 
  rel_cc_on_propForm E G (andp a (andp b c)) (andp (andp a b) c).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply andb_true_iff in H0;destruct H0.
apply andb_true_iff in H1;destruct H1.
apply andb_true_iff;split;auto.
apply andb_true_iff;split;auto.
Qed.

Lemma AS2_sound: forall E:_, forall a b c:_, forall G:_, 
  rel_cc_on_propForm E G (andp (andp a b) c) (andp a (andp b c)).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply andb_true_iff in H0;destruct H0.
apply andb_true_iff in H0;destruct H0.
apply andb_true_iff;split;auto.
apply andb_true_iff;split;auto.
Qed.

Lemma AS3_sound: forall E:_, forall a b c:_, forall G:_, 
  rel_cc_on_propForm E G (orp a (orp b c)) (orp (orp a b) c).
Proof.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply orb_true_iff in H0;destruct H0.
apply orb_true_iff;left;apply orb_true_iff;auto.
apply orb_true_iff in H0;destruct H0.
apply orb_true_iff;left;apply orb_true_iff;auto.
apply orb_true_iff;right;auto.
Qed.

Lemma AS4_sound: forall E:_, forall a b c:_, forall G:_, 
  rel_cc_on_propForm E G (orp (orp a b) c) (orp a (orp b c)).
Proof.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply orb_true_iff in H0;destruct H0.
apply orb_true_iff in H0;destruct H0.
apply orb_true_iff;left;auto.
apply orb_true_iff;right;apply orb_true_iff;auto.
apply orb_true_iff;right;apply orb_true_iff;auto.
Qed.

Lemma COM1_sound: forall E:_, forall a b:_, forall G:_, 
  rel_cc_on_propForm E G (andp a b) (andp b a).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply andb_true_iff in H0;destruct H0.
apply andb_true_iff;split;auto.
Qed.

Lemma COM2_sound: forall E:_, forall a b:_, forall G:_, 
  rel_cc_on_propForm E G (orp a b) (orp b a).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply orb_true_iff in H0;destruct H0;apply orb_true_iff;auto.
Qed.

Lemma DS1_sound: forall E:_, forall a b c:_, forall G:_,
  rel_cc_on_propForm E G (orp a (andp b c)) (andp (orp a b) (orp a c)).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply orb_true_iff in H0;destruct H0.
apply andb_true_iff;split;apply orb_true_iff;auto.
apply andb_true_iff in H0;destruct H0;apply andb_true_iff;split;apply orb_true_iff;auto.
Qed.

Lemma DS2_sound: forall E:_, forall a b c:_, forall G:_,
  rel_cc_on_propForm E G (andp (orp a b) (orp a c)) (orp a (andp b c)).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply andb_true_iff in H0;destruct H0.
apply orb_true_iff in H0;apply orb_true_iff in H1;destruct H0;destruct H1;apply orb_true_iff;auto.
right;apply andb_true_iff;split;auto.
Qed.

Lemma DS3_sound: forall E:_, forall a b c:_, forall G:_,
  rel_cc_on_propForm E G (andp a (orp b c)) (orp (andp a b) (andp a c)).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply andb_true_iff in H0;destruct H0.
apply orb_true_iff in H1;destruct H1.
apply orb_true_iff;left;apply andb_true_iff;split;auto.
apply orb_true_iff;right;apply andb_true_iff;split;auto.
Qed.

Lemma DS4_sound: forall E:_, forall a b c:_, forall G:_,
  rel_cc_on_propForm E G (orp (andp a b) (andp a c)) (andp a (orp b c)).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply orb_true_iff in H0;destruct H0;apply andb_true_iff in H0;destruct H0;apply andb_true_iff;split;auto;apply orb_true_iff;auto.
Qed.

Lemma AB1_sound: forall E:_, forall a b: propForm, forall G:_,
  rel_cc_on_propForm E G (orp a (andp a b)) a.
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply orb_true_iff in H0;destruct H0;auto;apply andb_true_iff in H0;destruct H0;auto.
Qed.

Lemma AB2_sound: forall E:_, forall a b: propForm, forall G:_,
  rel_cc_on_propForm E G a (orp a (andp a b)).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply orb_true_iff;auto.
Qed.

Lemma ID1_sound: forall E:_, forall a: propForm, forall G:_,
  rel_cc_on_propForm E G a (andp a a).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply andb_true_iff;split;auto.
Qed.

Lemma ID2_sound: forall E:_, forall a: propForm, forall G:_,
  rel_cc_on_propForm E G (orp a a) a.
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply orb_true_iff in H0;destruct H0;auto.
Qed.

Lemma SUP_sound: forall E:_, forall a b:_, forall G:_,
  rel_cc_on_propForm E G a (orp a b).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply orb_true_iff;auto.
Qed.

Lemma INF_sound: forall E:_, forall a b:_, forall G:_,
  rel_cc_on_propForm E G (andp a b) a.
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply andb_true_iff in H0;destruct H0;auto.
Qed.

Lemma TOP_sound: forall E:_, forall a:_, forall G:_,
  rel_cc_on_propForm E G a (andp a top).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply andb_true_iff;split;simpl;auto.
Qed.

Lemma BOT_sound: forall E:_, forall a:_, forall G:_,
  rel_cc_on_propForm E G (orp a bot) a.
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros.
apply orb_true_iff in H0;destruct H0;auto;inversion H0.
Qed.

Lemma CTX_sound: forall E:_, forall a b c:_, forall x:propVar, 
  forall G:_, (rel_cc_on_propForm E G a b) -> 
    (rel_cc_on_propForm E G (replace c x a) (replace c x b)).
Proof.
intros.
destruct H as [R H].
destruct H.
destruct H0.
destruct H1.
exists R;split;auto.
split;auto.
split;auto.
intro;intros.
induction c;simpl in *;auto.
destruct (eq_dec_nat x p).
rewrite e in *.
rewrite<- (EqNat.beq_nat_refl p) in *;auto.
assert (EqNat.beq_nat x p=false).
  apply EqNat.beq_nat_false_iff.
  intro;auto.
rewrite H5 in *;auto.
apply orb_true_iff;apply orb_true_iff in H4;destruct H4.
left;apply IHc1;auto.
right;apply IHc2;auto.
apply andb_true_iff;apply andb_true_iff in H4;destruct H4;split.
apply IHc1;auto.
apply IHc2;auto.
Qed.

Lemma TRA_sound: forall E:_, forall a b c:_, forall G:_, 
    (((rel_cc_on_propForm E G a b) /\ (rel_cc_on_propForm E G b c)) ->
      (rel_cc_on_propForm E G a c)).
Proof.
intros.
destruct H.
apply (rel_cc_trans E G a b c);auto.
Qed.

Lemma REF_sound: forall E:_, forall a:_, forall G:_, 
  (rel_cc_on_propForm E G a a).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H.
split.
apply empty_relation_rel_cc.
intro;intros;auto.
Qed.

Lemma CNT_sound: forall E:_, forall x y:_, forall G:relation propVar,
  ((G (x,y)) -> (rel_cc_on_propForm E G (var x) (var y))).
Proof.
intros.
exists (Empty_set _).
split.
apply Empty_is_finite.
split.
intro;intros;inversion H0.
split.
apply empty_relation_rel_cc.
intro;intros.
apply (H0 x);auto.
apply Union_intror;auto.
Qed.

Lemma CC_sound : forall E:_, forall x y:propVar, forall G:_,
  (bnd E x) /\ (bnd E y) /\ (bnd_block_number E x=bnd_block_number E y) ->
    (rel_cc_on_propForm E (Union _ G (Singleton _ (x,y))) (rhs E x) (rhs E y)) -> (rel_cc_on_propForm E G (var x) (var y)).
Proof.
intros.
destruct H;destruct H1.
destruct H0 as [R H0].
destruct H0.
destruct H3.
destruct H4.
assert (relative_cc E (Union _ R (Singleton _ (x,y))) G).
  split;auto.
  intro;intros.
  destruct H6.
  apply H4;auto.
  destruct x0;simpl in *.
  assert ((x,y)=(p,p0)).
    elim H6;auto.
  assert (x=p).
    congruence.
  assert (y=p0).
    congruence.
  rewrite H9 in *;rewrite H8 in *;auto.
  intro;intros.
  intro;intros.
  destruct (consistent_environment_rel_union theta _ _ H7).
  destruct (consistent_environment_rel_union theta _ _ H9).
  destruct H6.
  apply H4;auto.
  apply rel_union_consistent_environment;split;auto.
  apply rel_union_consistent_environment;split;auto.
  assert ((x,y)=x0).
    elim H6;auto.
  destruct x0;simpl in *.
  assert (x=p).
    congruence.
  assert (y=p0).
    congruence.
  rewrite H15 in *;rewrite H14 in *;apply H5;auto.
  apply rel_union_consistent_environment;split;auto.
  apply rel_union_consistent_environment;split;auto.
exists (Union _ R (Singleton _ (x,y))).
split.
apply finite_union;auto.
apply (eq_dec_pairs _ eq_dec_nat).
rewrite (Extensionality_Ensembles _ (Singleton (propVar * propVar) (x, y)) (Add _ (Empty_set _) (x,y))).
apply Union_is_finite.
apply Empty_is_finite.
intro;inversion H7.
split;intro;intros.
right;auto.
destruct H7;auto.
inversion H7.
split.
intro;intros.
destruct H7.
apply H3;auto.
destruct x0;simpl in *.
assert ((x,y)=(p,p0)).
  elim H7;auto.
assert (x=p).
  congruence.
assert (y=p0).
  congruence.
rewrite H9 in *;rewrite H10 in *;auto.
split;auto.
intro;intros.
apply (H7 x);auto.
apply Union_introl;apply Union_intror;apply In_singleton.
Qed.

Definition mk_rel_cc(E:_)(s : statement) : Prop :=
match s with
| stmt G f g => rel_cc_on_propForm E G f g
end.

Lemma soundness: forall E:_, forall f g:_, forall G:_, 
  prv_tree E (stmt G f g) -> mk_rel_cc E (stmt G f g).
Proof.
intros.
induction H.
apply AS1_sound.
apply AS2_sound.
apply AS3_sound.
apply AS4_sound.
apply COM1_sound.
apply COM2_sound.
apply DS1_sound.
apply DS2_sound.
apply DS3_sound.
apply DS4_sound.
apply AB1_sound.
apply AB2_sound.
apply ID1_sound.
apply ID2_sound.
apply SUP_sound.
apply INF_sound.
apply TOP_sound.
apply BOT_sound.
apply CTX_sound.
assumption.
apply (TRA_sound E a b c).
split;assumption.
apply REF_sound.
apply (CC_sound).
auto.
unfold mk_rel_cc in IHprv_tree.
assumption.
apply CNT_sound.
assumption.
Qed.

Lemma prv_system_sound : forall E:_, forall x y:propVar,
  prv_tree E (stmt (Empty_set _) (var x) (var y)) -> ((cc_max E) (x,y) \/ x=y).
Proof.
intros.
destruct (eq_dec_nat x y);auto.
left.
specialize (soundness _ _ _ _ H).
intro.
destruct H0 as [R H0].
destruct H0.
destruct H1.
destruct H2.
exists (reachable _ R).
split.
apply empty_relative_cc.
apply rel_cc_reach;auto.
split.
clear - H1.
intro;intros;induction H;auto.
destruct (H1 _ H0);destruct IHreachable;simpl in *;split;auto.

destruct (exists_list _ _ H0) as [l H4].
destruct (reachable_from_cons _ _ H4 x (var y)).
assert (rel_cons R (var x) (var y)).
clear - H3.
intro;intros.
apply H3;auto.
intro;intros.
apply (H x0);auto.
destruct H1;auto.
inversion H1.
specialize (H6 H7).
unfold reachable_from in H6.
simpl in H6.
destruct (reachable_dec_b nat eq_dec_nat l R H4 x y);auto.
exfalso.
destruct (eq_dec_nat x y);auto.
inversion H6.
Qed.

Lemma prv_system_sound_bnd : forall E:_, forall x y:_, bnd E x -> bnd E y ->
  prv_tree E (stmt (Empty_set _) (var x) (var y)) -> (cc_max E (x,y)).
Proof.
intros.
destruct (prv_system_sound _ _ _ H1);auto.
exists (Union _ (Empty_set _) (Singleton _ (x, y))).
rewrite H2 in *.
remember (y,y) as XY.
split.
split.
intro;intros.
destruct H3.
inversion H3.
destruct H3.
rewrite HeqXY in *;simpl;auto.
intro;intros.
destruct H3.
inversion H3.
intro;intros.
destruct H3.
rewrite HeqXY in *;simpl in *;auto.
split.
intro;intros.
destruct H3.
inversion H3.
destruct H3.
rewrite HeqXY in *;simpl;auto.
apply Union_intror.
apply In_singleton.
Qed.
