Add LoadPath projectPath.

Require Import propositional_formulas.
Require Import propositional_formulas_semantics.
Require Import consequence_propositional_formula.
Require Import Bool.

(*----------DNF----------*)

Definition singleton_propForm (f:propForm):Prop :=
match f with
| top => True
| bot => True
| var x => True
| _ => False
end.

Fixpoint conj_propForm (f:propForm):Prop :=
match f with
| a /\p b => conj_propForm a /\ conj_propForm b
| _ => singleton_propForm f
end.

Definition conj_ind : forall P:propForm -> Prop,
  P top ->
  P bot ->
  (forall x:propVar, P (var x)) ->
  (forall f g:propForm, conj_propForm f -> conj_propForm g -> P f -> P g ->
    P (f /\p g)) ->
  (forall f:propForm, conj_propForm f -> P f).
Proof.
intros;induction f;simpl in *;auto.
inversion H3.
destruct H3;auto.
Qed.

Fixpoint DNF_propForm (f:propForm):Prop := 
match f with
| f1 \/p f2 => (DNF_propForm f1 /\ DNF_propForm f2)
| _ => conj_propForm f
end.

Definition DNF_ind : forall P:propForm -> Prop,
  P top ->
  P bot ->
  (forall x:propVar, P (var x)) ->
  (forall f g:propForm, conj_propForm f -> conj_propForm g -> P f -> P g -> P (f /\p g)) ->
  (forall f g:propForm, DNF_propForm f -> DNF_propForm g -> P f -> P g -> P (f \/p g)) ->
  (forall f:propForm, DNF_propForm f -> P f).
Proof.
intros.
induction f;simpl in *;auto.
destruct H4;apply H3;auto.
destruct H4;apply H2;auto.
apply IHf1.
destruct f1;unfold DNF_propForm;auto.
inversion H4.
destruct f2;unfold DNF_propForm;auto.
inversion H5.
Qed.

Fixpoint distribute (f:propForm) : propForm -> propForm :=
fix distribute1 (g:propForm) : propForm :=
match f with
| f1 \/p f2 => distribute f1 g \/p distribute f2 g
| _ => match g with
       | g1 \/p g2 => distribute1 g1 \/p distribute1 g2
       | _ => f /\p g
       end
end.

Lemma distribute_DNF : forall f g:propForm, DNF_propForm f -> DNF_propForm g -> 
  DNF_propForm (distribute f g).
Proof.
intros.
induction f;simpl in *;auto.
induction g;simpl in *;auto.
  destruct H0;split;auto.

induction g;simpl in *;auto.
  destruct H0;split;auto.

induction g;simpl in *;auto.
  destruct H0;split;auto.

destruct H.
specialize (IHf1 H).
specialize (IHf2 H1).
induction g;simpl in *;auto.

destruct H.
assert (DNF_propForm f1).
  destruct f1;simpl in *;auto.
  inversion H.
assert (DNF_propForm f2).
  destruct f2;simpl in *;auto.
  inversion H1.
specialize (IHf1 H2).
specialize (IHf2 H3).
induction g;simpl in *;auto.
  destruct H0.
  split;auto.
  apply IHg1;auto.
  destruct f1;destruct IHf1;simpl in *;auto.
  inversion H.
  destruct f2;destruct IHf2;simpl in *;auto.
  inversion H1.
  apply IHg2;auto.
  destruct f1;destruct IHf1;simpl in *;auto.
  inversion H.
  destruct f2;destruct IHf2;simpl in *;auto.
  inversion H1.
Qed.

Lemma distribute_sound : forall f:propForm, DNF_propForm f -> forall g:propForm, DNF_propForm g ->
  propForm_eqv (f /\p g) (distribute f g).
Proof.
intros f H;induction H using DNF_ind.
intros g H;induction H using DNF_ind;split;intro;auto;apply orb_true_iff;apply orb_true_iff in H1;destruct H1.
  left;apply IHDNF_propForm1;simpl;auto.
  right;apply IHDNF_propForm2;simpl;auto.
  left;apply IHDNF_propForm1;simpl;auto.
  right;apply IHDNF_propForm2;simpl;auto.

intros g H;induction H using DNF_ind;split;intro;auto.
  simpl in H1;inversion H1.
  apply orb_true_iff in H1;destruct H1.
  destruct (IHDNF_propForm1 theta);simpl in *;auto.
  destruct (IHDNF_propForm2 theta);simpl in *;auto.

intros g H;induction H using DNF_ind;split;intro;auto.
  apply andb_true_iff in H1;destruct H1.
  apply orb_true_iff in H2;apply orb_true_iff;destruct H2.
  left;apply IHDNF_propForm1;auto;apply andb_true_iff;split;auto.
  right;apply IHDNF_propForm2;auto;apply andb_true_iff;split;auto.

  apply orb_true_iff in H1;destruct H1.
  assert ([[var x /\p f]]theta).
    apply IHDNF_propForm1;auto.
  apply andb_true_iff in H2;destruct H2;apply andb_true_iff;split;auto.
  apply orb_true_iff;auto.
  assert ([[var x /\p g]]theta).
    apply IHDNF_propForm2;auto.
  apply andb_true_iff in H2;destruct H2;apply andb_true_iff;split;auto.
  apply orb_true_iff;auto.

intros g0 H1;induction H1 using DNF_ind;split;intro;auto.
  apply andb_true_iff in H1;destruct H1;apply andb_true_iff in H1;destruct H1.
  apply orb_true_iff;apply orb_true_iff in H2;destruct H2.
  left;apply IHDNF_propForm1;apply andb_true_iff;split;auto;apply andb_true_iff;split;auto.
  right;apply IHDNF_propForm2;apply andb_true_iff;split;auto;apply andb_true_iff;split;auto.

  apply orb_true_iff in H1.
  destruct H1.
  destruct (IHDNF_propForm1 theta).
  specialize (H3 H1).
  apply andb_true_iff in H3;destruct H3;apply andb_true_iff in H3;destruct H3.
  try repeat (apply andb_true_iff;split);try apply orb_true_iff;auto.
  destruct (IHDNF_propForm2 theta).
  specialize (H3 H1).
  apply andb_true_iff in H3;destruct H3;apply andb_true_iff in H3;destruct H3.
  try repeat (apply andb_true_iff;split);try apply orb_true_iff;auto.

split;intro.
  apply andb_true_iff in H2;destruct H2.
  apply orb_true_iff in H2;destruct H2.
  destruct (IHDNF_propForm1 g0 H1 theta).
  assert ([[distribute f g0]]theta).
    apply H4;apply andb_true_iff;split;auto.
  induction H1 using DNF_ind;apply orb_true_iff;auto.
  destruct (IHDNF_propForm2 g0 H1 theta).
  assert ([[distribute g g0]]theta).
    apply H4;apply andb_true_iff;split;auto.
  induction H1 using DNF_ind;apply orb_true_iff;auto.

  assert ([[distribute f g0]]theta \/ [[distribute g g0]]theta).
    destruct g0;apply orb_true_iff in H2;destruct H2;auto.
  destruct (IHDNF_propForm1 _ H1 theta).
  destruct (IHDNF_propForm2 _ H1 theta).
  destruct H3.
  specialize (H5 H3).
  apply andb_true_iff in H5;destruct H5;apply andb_true_iff;split;auto;apply orb_true_iff;auto.
  specialize (H7 H3).
  apply andb_true_iff in H7;destruct H7;apply andb_true_iff;split;auto;apply orb_true_iff;auto.
Qed.

Fixpoint toDNF (f:propForm):propForm :=
match f with
| f1 \/p f2 => toDNF f1 \/p toDNF f2
| f1 /\p f2 => distribute (toDNF f1) (toDNF f2)
| _ => f
end.

Lemma DNF_toDNF : forall f:propForm, DNF_propForm (toDNF f).
Proof.
intros.
induction f;simpl in *;auto.
apply distribute_DNF;auto.
Qed.

Lemma toDNF_eqv : forall f:propForm, propForm_eqv f (toDNF f).
Proof.
intro;induction f;split;intros;simpl in *;auto.
apply orb_true_iff in H;apply orb_true_iff;destruct H.
left;apply IHf1;auto.
right;apply IHf2;auto.
apply orb_true_iff in H;apply orb_true_iff;destruct H.
left;apply IHf1;auto.
right;apply IHf2;auto.
apply distribute_sound;try apply DNF_toDNF;auto.
apply andb_true_iff in H;apply andb_true_iff;destruct H;split.
apply IHf1;auto.
apply IHf2;auto.
apply distribute_sound in H;try apply DNF_toDNF;auto.
apply andb_true_iff in H;apply andb_true_iff;destruct H;split.
apply IHf1;auto.
apply IHf2;auto.
Qed.









































