Add LoadPath projectPath.

Require Import propositional_formulas.
Require Import propositional_formulas_semantics.
Require Import consequence_propositional_formula.
Require Import propositional_formula_DNF.
Require Import Bool.
Require Import Arith.

Lemma conj_used_cons : forall f:propForm, conj_propForm f -> forall x:propVar, uses f x -> propForm_cons f (var x).
Proof.
intros f H;induction H using conj_ind.
intros;inversion H.
intros;inversion H.
intros.
apply beq_nat_true_iff in H.
rewrite H;intro;intros;auto.
intros.
intro;intros.
apply andb_true_iff in H2;destruct H2.
apply orb_true_iff in H1;destruct H1.
apply IHconj_propForm1;auto.
apply IHconj_propForm2;auto.
Qed.

Lemma conj_unused : forall f:propForm, conj_propForm f -> ~propForm_eqv f bot -> 
  forall x:propVar, ~uses f x -> [[f]]environment_max_point x.
Proof.
intros f H;induction H using conj_ind;intros.
simpl;auto.

exfalso;apply H;split;intro;auto.

simpl in H0.
simpl.
unfold environment_max_point.
assert (beq_nat x0 x = false).
  apply beq_nat_false_iff;intro;apply H0;apply Nat.eqb_eq;auto.
rewrite H1;auto.

apply andb_true_iff;split.
apply IHconj_propForm1;auto.
intro.
apply H1.
split;intro.
apply andb_true_iff in H4;apply H3;destruct H4;auto.
inversion H4.
intro;apply H2;apply orb_true_iff;auto.
apply IHconj_propForm2;auto.
intro.
apply H1.
split;intro.
apply andb_true_iff in H4;apply H3;destruct H4;auto.
inversion H4.
intro;apply H2;apply orb_true_iff;auto.
Qed.

Lemma conj_environment_intersect : forall f:propForm, conj_propForm f ->
  forall theta1 theta2:environment, [[f]]theta1 -> [[f]]theta2 -> [[f]]environment_intersect theta1 theta2.
Proof.
intros f H;induction H using conj_ind;intros;auto.
unfold environment_intersect;apply andb_true_iff;split;auto.

apply andb_true_iff in H1;destruct H1;apply andb_true_iff in H2;destruct H2.
apply andb_true_iff;split.
apply IHconj_propForm1;auto.
apply IHconj_propForm2;auto.
Qed.

Lemma conj_cons_propForm_decidable : forall f:propForm, forall g:propForm, conj_propForm f ->
  (propForm_cons f g \/ exists theta:environment, ([[f]]theta /\ ~[[g]]theta)).
Proof.
intros;induction g.
left;intro;intros;simpl;auto.

destruct (exists_env_dec f).
destruct H0 as [theta H0];right;exists theta;split;auto.
intro;inversion H1.
left;intro;intros.
exfalso;apply (H0 theta);auto.

destruct (used_or_not f p).
left;apply conj_used_cons;auto.
destruct (exists_env_dec f).
right.
exists (environment_max_point p).
split.
apply conj_unused;auto.
intro.
destruct H1 as [theta H1];destruct (H2 theta).
specialize (H3 H1);inversion H3.
unfold environment_max_point.
simpl;intro.
rewrite<- beq_nat_refl in H2;inversion H2.
left;intro;intros.
exfalso;apply (H1 theta);auto.

destruct IHg1.
left;intro;intros.
apply orb_true_iff;left;apply H0;auto.
destruct IHg2.
left;intro;intros.
apply orb_true_iff;right;apply H1;auto.
destruct H0 as [theta1 H0];destruct H0.
destruct H1 as [theta2 H1];destruct H1.
right;exists (environment_intersect theta1 theta2).
split.
apply conj_environment_intersect;auto.
intro.
apply orb_true_iff in H4.
destruct H4.
apply (environment_intersect_maintains_neg _ _ H2 _ H4).
destruct (environment_intersect_commutes theta1 theta2 g2).
apply (environment_intersect_maintains_neg _ _ H3 _ (H5 H4)).

intros.
destruct IHg1.
destruct IHg2.
left;intro;intros;apply andb_true_iff;split.
apply H0;auto.
apply H1;auto.
destruct H1 as [theta H1];destruct H1.
right;exists theta;split;auto.
intro;apply andb_true_iff in H3;destruct H3;auto.
destruct H0 as [theta H0];destruct H0.
right;exists theta;split;auto.
intro;apply andb_true_iff in H2;destruct H2;auto.
Qed.

Lemma DNF_cons_propForm_decidable : forall f:propForm, DNF_propForm f -> forall g:propForm,
  (propForm_cons f g \/ exists theta:environment, ([[f]]theta /\ ~[[g]]theta)).
Proof.
intros;induction H using DNF_ind;intros.
apply (conj_cons_propForm_decidable top g);simpl;auto.
apply (conj_cons_propForm_decidable bot g);simpl;auto.
apply (conj_cons_propForm_decidable (var x) g);simpl;auto.
apply (conj_cons_propForm_decidable (f /\p g0) g);simpl;auto.
destruct IHDNF_propForm1.
destruct IHDNF_propForm2.
left;intro;intro.
apply orb_true_iff in H3;destruct H3.
apply H1;auto.
apply H2;auto.
destruct H2 as [theta H2];destruct H2.
right.
exists theta;split;auto.
apply orb_true_iff;right;auto.
destruct H1 as [theta H1];destruct H1.
right.
exists theta;split;auto.
apply orb_true_iff;left;auto.
Qed.

Lemma propForm_cons_dec : forall f g:propForm, propForm_cons f g \/ exists theta:environment, [[f]]theta /\~[[g]]theta.
Proof.
intros.
destruct (DNF_cons_propForm_decidable (toDNF f) (DNF_toDNF f) g).
left;intro;intros.
apply H.
destruct (toDNF_eqv f theta);auto.
destruct H as [theta H];destruct H.
right;exists theta;split;auto.
apply (toDNF_eqv f theta);auto.
Qed.

Lemma propForm_eqv_decidable : forall f g:propForm, propForm_eqv f g \/ ~propForm_eqv f g.
Proof.
intros.
destruct (propForm_cons_dec f g).
destruct (propForm_cons_dec g f).
left;split;intro.
apply H;auto.
apply H0;auto.
destruct H0 as [theta H0];destruct H0.
right;intro.
apply H1;apply H2;auto.
destruct H as [theta H];destruct H.
right;intro.
apply H0;apply H1;auto.
Qed.

Lemma conj_uses_cons : forall f:propForm, conj_propForm f -> ~propForm_eqv f bot -> forall x:propVar,
  uses f x <-> propForm_cons f (var x).
Proof.
intros;induction H using conj_ind;split;intros;auto.
inversion H.
simpl;apply (H empty_environment);auto.
inversion H.
exfalso;apply H0;split;auto.
intro;intros.
unfold uses in H.
rewrite (EqNat.beq_nat_true _ _ H) in *;auto.
simpl.
apply (H (environment_point x0)).
unfold environment_point;simpl;rewrite<- EqNat.beq_nat_refl;auto.

intro;intros.
simpl in H2;apply orb_true_iff in H2;destruct H2.
apply IHconj_propForm1;auto.
intro.
apply H0.
split;intro.
apply H4;apply andb_true_iff in H5;destruct H5;auto.
inversion H5.
apply andb_true_iff in H3;destruct H3;auto.
apply IHconj_propForm2;auto.
intro.
apply H0.
split;intro.
apply H4;apply andb_true_iff in H5;destruct H5;auto.
inversion H5.
apply andb_true_iff in H3;destruct H3;auto.

assert ([[var x]]min_environment(f /\p g)).
  apply H2.
  apply propForm_min_env;auto.
simpl in H3;auto.
Qed.

Lemma min_env_cons : forall f g:propForm, conj_propForm f -> ~propForm_eqv f bot ->
  [[g]]min_environment f <-> propForm_cons f g.
Proof.
intros;induction g;split;intros;auto.
intro;intros;auto.
inversion H1.

apply H1.
apply propForm_min_env;auto.
intro;intros.
unfold min_environment in H1;simpl in H1.
apply (conj_uses_cons f);auto.
apply (conj_uses_cons f);auto.

intro;intros.
simpl in *;apply orb_true_iff;apply orb_true_iff in H1;destruct H1.
left;apply IHg1;auto.
right;apply IHg2;auto.

apply H1;apply propForm_min_env;auto.
intro;intros.
simpl in *;apply andb_true_iff;apply andb_true_iff in H1;destruct H1;split.
apply IHg1;auto.
apply IHg2;auto.

apply H1;apply propForm_min_env;auto.
Qed.


