Add LoadPath projectPath.

Require Import Bool.
Require Import Arith.
Require Import relation_definitions.


Coercion is_true : bool >-> Sortclass.

(*----------propForm----------*)

Definition propVar:=nat.

Inductive propForm : Set :=
| top : propForm
| bot : propForm
| var : propVar -> propForm
| orp : propForm -> propForm -> propForm
| andp : propForm -> propForm -> propForm.

Lemma eq_dec_propForm : eq_dec propForm.
Proof.
intro.
induction x;intros.
induction y;auto;right;intro;inversion H.
induction y;auto;right;intro;inversion H.
induction y;auto.
right;intro;inversion H.
right;intro;inversion H.
destruct (eq_dec_nat p p0).
rewrite e;left;auto.
right;intro.
apply n;congruence.
right;intro;inversion H.
right;intro;inversion H.
induction y.
right;intro;inversion H.
right;intro;inversion H.
right;intro;inversion H.
destruct (IHx1 y1).
destruct (IHx2 y2).
rewrite e in *;rewrite e0 in *;auto.
right;intro.
apply n;congruence.
right;intro.
apply n;congruence.
right;intro;inversion H.
induction y.
right;intro;inversion H.
right;intro;inversion H.
right;intro;inversion H.
right;intro;inversion H.
destruct (IHx1 y1).
destruct (IHx2 y2).
rewrite e in *;rewrite e0 in *;auto.
right;intro.
apply n;congruence.
right;intro.
apply n;congruence.
Qed.

Notation "A \/p B" := (orp A B) (at level 85).
Notation "A /\p B" := (andp A B) (at level 80).

Fixpoint uses (f:propForm)(x:propVar) : bool :=
match f with 
| a \/p b => (uses a x) || (uses b x)
| a /\p b => (uses a x) || (uses b x)
| var y => (beq_nat y x)
| top => false
| bot => false
end.

Lemma used_or_not : forall f:propForm, forall x:propVar, uses f x \/ ~uses f x.
Proof.
intros.
destruct (uses f x);auto.
right;intro;inversion H.
Qed.


Fixpoint replace (f:propForm)(x:propVar)(g:propForm):propForm :=
match f with
| a /\p b => (replace a x g) /\p (replace b x g)
| a \/p b => (replace a x g) \/p (replace b x g)
| var y => if (beq_nat x y) then g else var y
| _ => f
end.

Notation "f [ x := f' ]" := (replace f x f')(at level 50).

(*for every propForm there exists a variable which is not used in this propForm*)
Fixpoint mxUsed(f:propForm):propVar:=
match f with
| var x => x
| g1 /\p g2 => max (mxUsed g1) (mxUsed g2)
| g1 \/p g2 => max (mxUsed g1) (mxUsed g2)
| _ => 0
end.

Lemma maxlteq : forall x y z:nat, (x<=y\/x<=z)->x<=(max y z).
Proof.
intros.
destruct (Nat.max_spec y z).
destruct H0.
rewrite H1 in *.
destruct H;auto.
assert (y<=z).
  apply lt_le_weak;auto.
apply (le_trans _ y);auto.
destruct H0.
rewrite H1 in *.
destruct H;auto.
apply (le_trans _ z);auto.
Qed.

Lemma uses_lteq_mxUsed : forall f:propForm, forall x:propVar, uses f x -> x<=(mxUsed f).
Proof.
induction f;simpl in *;intros;auto.
inversion H.
inversion H.
apply beq_nat_true_iff in H.
rewrite H in *;auto.
apply maxlteq.
apply orb_true_iff in H.
destruct H;auto.
apply maxlteq.
apply orb_true_iff in H.
destruct H;auto.
Qed.

Lemma suc_mxUsed_unused : forall f:propForm, ~uses f (S (mxUsed f)).
Proof.
induction f;simpl;intros;intro;auto.
inversion H.
inversion H.
induction p;simpl.
inversion H.
apply beq_nat_true_iff in H.
apply (n_Sn p);auto.
apply orb_true_iff in H.
rewrite Nat.succ_max_distr in H.
destruct (Nat.max_spec (S (mxUsed f1)) (S (mxUsed f2))).
destruct H0.
rewrite H1 in *.
destruct H.
clear H1.
specialize (uses_lteq_mxUsed _ _ H).
intro.
apply le_S_gt in H1.
apply lt_n_Sm_le in H0.
apply le_S_gt in H0.
apply (gt_asym _ _ H0);auto.
apply IHf2;auto.
destruct H0.
rewrite H1 in *.
destruct H.
apply IHf1;auto.
clear H1.
specialize (uses_lteq_mxUsed _ _ H).
intro.
apply le_S_gt in H1.
apply lt_n_Sm_le in H0.
apply (le_not_gt _ _ H0);auto.

apply orb_true_iff in H.
rewrite Nat.succ_max_distr in H.
destruct (Nat.max_spec (S (mxUsed f1)) (S (mxUsed f2))).
destruct H0.
rewrite H1 in *.
destruct H.
clear H1.
specialize (uses_lteq_mxUsed _ _ H).
intro.
apply le_S_gt in H1.
apply lt_n_Sm_le in H0.
apply le_S_gt in H0.
apply (gt_asym _ _ H0);auto.
apply IHf2;auto.
destruct H0.
rewrite H1 in *.
destruct H.
apply IHf1;auto.
clear H1.
specialize (uses_lteq_mxUsed _ _ H).
intro.
apply le_S_gt in H1.
apply lt_n_Sm_le in H0.
apply (le_not_gt _ _ H0);auto.
Qed.

Lemma exists_unused : forall f:propForm, exists x:propVar, ~uses f x.
Proof.
intros.
exists (S (mxUsed f)).
apply suc_mxUsed_unused.
Qed.

Lemma unused_no_rewrite : forall f f':propForm, forall x:propVar, (~uses f x) -> f[x:=f']=f.
Proof.
intros.
induction f;simpl in *.
auto.
auto.
assert (beq_nat x p=false).
apply beq_nat_false_iff.
intro.
apply H;apply beq_nat_true_iff;auto.
rewrite H0.
auto.
assert (~uses f1 x /\ ~uses f2 x).
split;intro;apply H;apply orb_true_iff;auto.
destruct H0.
specialize (IHf1 H0).
specialize (IHf2 H1).
rewrite IHf1.
rewrite IHf2.
auto.
assert (~uses f1 x /\ ~uses f2 x).
split;intro;apply H;apply orb_true_iff;auto.
destruct H0.
specialize (IHf1 H0).
specialize (IHf2 H1).
rewrite IHf1.
rewrite IHf2.
auto.
Qed.


