Add LoadPath projectPath.

Require Import Bool.
Require Import Arith.
Require Import propositional_formulas.


(*--------propForm semantics--------*)
Definition environment := propVar -> bool.

Definition redefineEnvironment (theta:environment)(x:propVar)(s:bool) : environment :=
fun (y:propVar) => if beq_nat x y then s else theta y.

Definition full_environment (x:propVar) : bool := true.

Definition empty_environment (x:propVar) : bool := false.

Definition environment_point (x:propVar) : environment := fun y => beq_nat x y.

Definition environment_max_point (x y:propVar) := if beq_nat x y then false else true.

Notation "theta [ x := f ]" := (redefineEnvironment theta x f) (at level 50).

Definition environment_union (theta1 theta2:environment)(x:propVar) := (theta1 x) || (theta2 x).

Definition environment_intersect (theta1 theta2:environment)(x:propVar) := (theta1 x) && (theta2 x).

Fixpoint propForm_solution (f:propForm)(theta:environment) : bool :=
match f with
| a \/p b => (propForm_solution a theta) || (propForm_solution b theta)
| a /\p b => (propForm_solution a theta) && (propForm_solution b theta)
| var x => theta x
| top => true
| bot => false
end.

Notation "[[ f ]] theta" := (propForm_solution f theta) (at level 30).

Definition min_environment (f:propForm):environment :=
fun (x:propVar) => uses f x.

Lemma propForm_solution_choice : forall f:propForm, forall theta:environment,
  [[f]]theta \/ ~[[f]]theta.
Proof.
intros;destruct ([[f]]theta);auto.
right;intro;inversion H.
Qed.

Lemma environment_union_maintains: forall theta1:environment, forall f1:propForm,
  [[f1]]theta1 -> forall theta2:environment, [[f1]](environment_union theta1 theta2).
Proof.
induction f1;simpl;intros;auto.
apply orb_true_iff;left;auto.
apply orb_true_iff in H.
apply orb_true_iff.
destruct H.
left;apply IHf1_1;auto.
right;apply IHf1_2;auto.
apply andb_true_iff in H.
apply andb_true_iff.
destruct H.
split.
apply IHf1_1;auto.
apply IHf1_2;auto.
Qed.

Lemma environment_union_commutes: forall theta1 theta2:environment, forall f:propForm,
  [[f]](environment_union theta1 theta2) <-> [[f]](environment_union theta2 theta1).
Proof.
induction f;simpl;split;intro;auto.
apply orb_true_iff.
apply orb_true_iff in H;destruct H;auto.
apply orb_true_iff.
apply orb_true_iff in H;destruct H;auto.
apply orb_true_iff.
apply orb_true_iff in H;destruct H.
left;apply IHf1;auto.
right;apply IHf2;auto.
apply orb_true_iff.
apply orb_true_iff in H;destruct H.
left;apply IHf1;auto.
right;apply IHf2;auto.
apply andb_true_iff in H.
apply andb_true_iff.
destruct H.
split.
apply IHf1;auto.
apply IHf2;auto.
apply andb_true_iff.
apply andb_true_iff in H.
destruct H.
split.
apply IHf1;auto.
apply IHf2;auto.
Qed.

Lemma exists_env_dec : forall f:propForm, (exists theta:environment, [[f]]theta) \/ forall theta:environment, ~[[f]]theta.
Proof.
induction f.
left;exists full_environment;simpl;auto.
right;intro;simpl;intro;inversion H.
left;exists full_environment;simpl;auto.
destruct IHf1.
destruct H as [theta H].
left;exists theta;apply orb_true_iff;left;auto.
destruct IHf2.
destruct H0 as [theta H0].
left;exists theta;apply orb_true_iff;right;auto.
right;intro;intro.
apply orb_true_iff in H1;destruct H1.
apply (H theta);auto.
apply (H0 theta);auto.

destruct IHf1.
destruct IHf2.
destruct H as [theta1 H].
destruct H0 as [theta2 H0].
left;exists (environment_union theta1 theta2);apply andb_true_iff;split.
apply environment_union_maintains;auto.
apply environment_union_commutes;apply environment_union_maintains;auto.

right;intros;intro;apply andb_true_iff in H1;destruct H1.
apply (H0 theta);auto.
right;intros;intro;apply andb_true_iff in H0;destruct H0.
apply (H theta);auto.
Qed.

Lemma environment_intersect_maintains_neg: forall theta1:environment, forall f1:propForm,
  ~[[f1]]theta1 -> forall theta2:environment, ~[[f1]](environment_intersect theta1 theta2).
Proof.
induction f1;simpl;intros;auto.
intro.
apply andb_true_iff in H0;destruct H0;auto.

intro.
assert (~[[f1_1]]theta1).
  intro;apply H;apply orb_true_iff;auto.
assert (~[[f1_2]]theta1).
  intro;apply H;apply orb_true_iff;auto.
apply orb_true_iff in H0;destruct H0.
apply (IHf1_1 H1 theta2);auto.
apply (IHf1_2 H2 theta2);auto.

intro.
apply andb_true_iff in H0;destruct H0.
destruct ([[f1_1]]theta1).
assert (~[[f1_2]]theta1).
  intro;apply H;apply andb_true_iff;split;auto.
apply (IHf1_2 H2 theta2);auto.
assert (~false).
  intro;inversion H2.
apply (IHf1_1 H2 theta2);auto.
Qed.

Lemma environment_intersect_commutes: forall theta1 theta2:environment, forall f:propForm,
  [[f]](environment_intersect theta1 theta2) <-> [[f]](environment_intersect theta2 theta1).
Proof.
induction f;simpl;split;intro;auto.
apply andb_true_iff.
apply andb_true_iff in H;destruct H;auto.
apply andb_true_iff.
apply andb_true_iff in H;destruct H;auto.
apply orb_true_iff.
apply orb_true_iff in H;destruct H.
left;apply IHf1;auto.
right;apply IHf2;auto.
apply orb_true_iff.
apply orb_true_iff in H;destruct H.
left;apply IHf1;auto.
right;apply IHf2;auto.
apply andb_true_iff in H.
apply andb_true_iff.
destruct H.
split.
apply IHf1;auto.
apply IHf2;auto.
apply andb_true_iff.
apply andb_true_iff in H.
destruct H.
split.
apply IHf1;auto.
apply IHf2;auto.
Qed.


