Add LoadPath projectPath.

Require Import Ensembles.
Require Import Finite_sets.

Require Import relation_definitions.
Require Import relation_lemmas.
Require Import transitive_closure_minimal_environment.

Lemma rem_is_Finite : forall A:Type, eq_dec A -> forall R:Ensemble A,
  Finite _ R -> forall x:_, Finite _ (Subtract _ R x).
Proof.
intros.
induction H.
rewrite<- (Extensionality_Ensembles _ (Empty_set _) (Subtract _ (Empty_set _) x)).
apply Empty_is_finite.
split;intro;intros.
inversion H.
destruct H;auto.

destruct (exists_list _ _ H) as [l H1].
destruct (List.in_dec X x0 l).
rewrite<- (Extensionality_Ensembles _ A0 (Add _ A0 x0));auto.
split;intro;intros.
left;auto.
destruct H2;auto.
destruct H2;apply H1;auto.

destruct (X x0 x).
rewrite e in *.
rewrite<- (Extensionality_Ensembles _ (Subtract A A0 x) (Subtract A (Add A A0 x) x));auto.
split;intro;intros.
destruct H2.
split;auto.
left;auto.
destruct H2.
destruct H2;split;auto.
contradiction.

rewrite<- (Extensionality_Ensembles _ (Add _ (Subtract A A0 x) x0) (Subtract A (Add A A0 x0) x));auto.
apply Union_is_finite;auto.
intro.
destruct H2.
apply n;apply H1;auto.

split;intro;intros.
destruct H2.
destruct H2;split;auto.
left;auto.
split;auto.
right;auto.
intro.
destruct H2;destruct H3;auto.
destruct H2.
destruct H2.
left;split;auto.
destruct H2.
right;apply In_singleton.
Qed.

Inductive rev_subset (A:Type)(R:Ensemble A):Ensemble A -> Prop :=
| full_rev_subset : rev_subset A R R
| rem_rev_subset : forall R1:Ensemble A, rev_subset A R R1 ->
  forall x:_, In _ R1 x -> rev_subset A R (Subtract _ R1 x).

Inductive rev_subset_card (A:Type)(R:Ensemble A): Ensemble A -> nat -> Prop :=
| card_full : rev_subset_card _ R R 0
| card_not_full : forall R1:_, forall n:nat, rev_subset_card _ R R1 n ->
  forall x:_, In _ R1 x -> rev_subset_card _ R (Subtract _ R1 x) (S n).

Lemma rev_subset_finite : forall A:Type, eq_dec A -> forall R R1:_,
  rev_subset A R R1 -> Finite _ R -> Finite _ R1.
Proof.
intros.
induction H;auto.

apply rem_is_Finite;auto.
Qed.

Lemma rev_subset_card0 : forall (A:_)(R1 R2:_)(n:_),
  rev_subset_card _ R1 R2 n -> n=0 -> Same_set (A*A) R1 R2.
Proof.
intros A R1 R2 n H;induction H;intros.
split;intro;auto.
inversion H1.
Qed.

Lemma rev_subset_card_finite : forall (A:_)(R1 R2:_)(n:_),
  eq_dec A -> Finite (A*A) R1 -> rev_subset_card _ R1 R2 n -> Finite _ R2.
Proof.
intros.
induction H0.
induction H.
apply Empty_is_finite.
apply Union_is_finite;auto.

apply rem_is_Finite;auto.
apply eq_dec_pairs;auto.
Qed.

Lemma rev_subset_card_rev_subset : forall A:_, forall e1 e2:_, forall n:_,
  rev_subset_card A e1 e2 n -> rev_subset A e1 e2.
Proof.
intros.
induction H.
apply full_rev_subset.
apply rem_rev_subset;auto.
Qed.

Lemma rev_subset_card_gt0 : forall A:_, forall e1 e2:_,
  forall n:_, rev_subset_card A e1 e2 n ->
    forall x:_, In _ e1 x -> ~In _ e2 x -> n>0.
Proof.
intros A e1 e2 n H;induction H;intros.
contradiction.
apply Gt.gt_Sn_O.
Qed.

Lemma rev_subset_card_minus : forall A:_, eq_dec A -> forall e1 e2:_, Finite _ e1 -> forall m:_,
  rev_subset_card A e1 e2 m -> forall x:_, In _ e1 x -> ~In _ e2 x ->
    rev_subset_card A e1 (Add _ e2 x) (m-1).
Proof.
intros A eq_decA e1 e2 H m H0;induction H0;intros.
contradiction.

simpl.
assert (n-0=n).
  destruct n;auto.
rewrite H4.
assert (Finite _ R1).
  apply (rev_subset_finite _ eq_decA e1);auto.
  apply (rev_subset_card_rev_subset _ _ _ _ H0).
destruct (Finite_dec _ eq_decA _ H5 x0).
assert (x=x0).
  destruct (eq_decA x x0);auto.
  exfalso.
  apply H3.
  split;auto.
  intro.
  apply n0;elim H7;auto.
rewrite H7 in *.
assert (Same_set _ R1 (Add _ (Subtract _ R1 x0) x0)).
  split;intro;intros.
  destruct (eq_decA x0 x1).
  rewrite e in *;right;apply In_singleton.
  left;split;auto.
  intro.
  apply n0;elim H9;auto.
  destruct H8.
  destruct H8;auto.
  elim H8;auto.
rewrite<- (Extensionality_Ensembles _ _ _ H8);auto.
assert (rev_subset_card A e1 (Add _ R1 x0) (n-1));auto.
assert (n>0).
  apply (rev_subset_card_gt0 _ e1 R1 _ H0 x0);auto.
assert (n=S (n-1)).
  clear - H8;induction n;simpl.
  inversion H8.
  destruct n;auto.
rewrite H9.
assert (Same_set _ (Add _ (Subtract _ R1 x) x0) (Subtract _ (Add _ R1 x0) x)).
  split;intro;intros.
  destruct H10.
  destruct H10.
  split;auto.
  left;auto.
  split.
  elim H10;right;apply In_singleton.
  intro.
  assert (x0=x1).
    elim H10;auto.
  assert (x=x1).
    elim H11;auto.
  rewrite H12 in *;rewrite H13 in *;auto.
  destruct H10.
  destruct H10.
  left.
  split;auto.
  right;auto.
rewrite (Extensionality_Ensembles _ _ _ H10).
apply card_not_full;auto.
left;auto.
Qed.

Lemma rev_subset_contains : forall A:_, forall e1 e2:_,
  rev_subset A e1 e2 -> Included _ e2 e1.
Proof.
intros.
induction H;intro;intros;auto.
destruct H1.
apply IHrev_subset;auto.
Qed.

Lemma exists_rev_subset : forall A:_, forall R R1:_, rev_subset A R R1 -> exists n:_, rev_subset_card _ R R1 n.
Proof.
intros.
induction H.
exists 0.
apply card_full.

destruct IHrev_subset as [n IH].
exists (S n).
apply card_not_full;auto.
Qed.

Lemma included_rev_subset_rev : forall A:_, eq_dec A -> forall e2:_, Finite A e2 -> forall e e1:_,
  Same_set _ e (Union _ e1 e2) -> Same_set _ (Intersection _ e1 e2) (Empty_set _) ->
    rev_subset A e e1.
Proof.
intros A X e2 H;induction H as [|e2];intros.
rewrite (Extensionality_Ensembles _ e1 e).
apply full_rev_subset.
split;intro;intros.
apply H;apply Union_introl;auto.
destruct H.
destruct (H _ H1);auto.
inversion H3.

assert (e x).
  apply H1;apply Union_intror;right;apply In_singleton.
assert (~e1 x).
  intro.
  assert (Empty_set _ x).
    apply H2.
    split;auto.
    right;apply In_singleton.
  inversion H5.
assert (Same_set _ e (Union _ (Add _ e1 x) e2)).
  split;intro;intros.
  destruct (X x x0).
  rewrite e0 in *.
  apply Union_introl;right;apply In_singleton.
  destruct H1.
  destruct (H1 _ H5).
  apply Union_introl;left;auto.
  destruct H7.
  apply Union_intror;auto.
  exfalso;apply n;elim H7;auto.
  apply H1.
  destruct H5.
  destruct H5.
  apply Union_introl;auto.
  apply Union_intror;right;auto.
  apply Union_intror;left;auto.
assert (Same_set _ (Intersection _ (Add _ e1 x) e2) (Empty_set _)).
  split;intro;intros.
  apply H2.
  destruct H6.
  destruct H6.
  split;auto.
  left;auto.
  assert (x=x0).
    elim H6;auto.
  exfalso;rewrite H8 in *;apply H0;auto.
  inversion H6.
specialize (IHFinite e _ H5 H6).
assert (Same_set _ e1 (Subtract _ (Add _ e1 x) x)).
  split;intro;intros.
  split.
  left;auto.
  intro.
  assert (x=x0).
    elim H8;auto.
  rewrite H9 in *;auto.
  destruct H7.
  destruct H7;auto.
  contradiction.
rewrite (Extensionality_Ensembles _ _ _ H7).
apply rem_rev_subset;auto.
right;apply In_singleton.
Qed.

Lemma empty_rev_subset : forall A:_, eq_dec A -> forall e:_, Finite _ e -> rev_subset A e (Empty_set _).
Proof.
intros.
apply (included_rev_subset_rev _ X _ H).
split;intro;intros.
apply Union_intror;auto.
destruct H0;auto.
inversion H0.

split;intro;intros.
destruct H0;auto.
inversion H0.
Qed.



















































