Add LoadPath projectPath.

Require Import Ensembles.
Require Import Finite_sets.
Require Import relation_definitions.

Lemma eq_dec_pairs : forall A:Type, eq_dec A -> eq_dec (prod A A).
Proof.
intros;intro;intros.
induction x.
induction y.
destruct (X a a0).
destruct (X b b0).
left;congruence.
right;intro;apply n;congruence.
right;intro;apply n;congruence.
Qed.

Lemma Finite_dec : forall (A:_), eq_dec A -> forall (R:_), Finite A R ->
  forall x:_, R x \/ ~R x.
Proof.
intros.
induction H.
right;intro;inversion H.

destruct IHFinite.
left;left;auto.
destruct (X x x0).
rewrite e in *;left;right;apply In_singleton.
right;intro.
destruct H2;auto.
apply n;elim H2;auto.
Qed.

Lemma Add_is_finite : forall A:_, eq_dec A -> forall R:_, Finite A R ->
  forall x:_, Finite _ (Add _ R x).
Proof.
intros.
induction H.
apply Union_is_finite.
apply Empty_is_finite.
intro;inversion H.
destruct (X x0 x).
rewrite e in *.
rewrite (Extensionality_Ensembles _ (Add _ (Add _ A0 x) x) (Add _ A0 x)).
apply Union_is_finite;auto.
split;intro;intros.
destruct H1;auto.
right;auto.
left;auto.

rewrite (Extensionality_Ensembles _ (Add _ (Add _ A0 x0) x) (Add _ (Add _ A0 x) x0)).
apply Union_is_finite;auto.
intro.
destruct H1;auto.
apply n;elim H1;auto.

split;intro;intros.
destruct H1.
destruct H1.
left;left;auto.
right;auto.
left;right;auto.
destruct H1.
destruct H1.
left;left;auto.
right;auto.
left;right;auto.
Qed.

Lemma finite_union : forall (A:_)(e1 e2:_), eq_dec A -> Finite A e1 -> Finite _ e2 ->
  Finite _ (Union _ e1 e2).
Proof.
intros.
induction H0.
rewrite (Extensionality_Ensembles _ (Union A e1 (Empty_set A)) e1);auto.
split;intro;intros.
destruct H0;auto;inversion H0.
apply Union_introl;auto.

rewrite (Extensionality_Ensembles _ (Union A e1 (Add A A0 x)) (Add _ (Union _ e1 A0) x));auto.
apply Add_is_finite;auto.
split;intro;intros.
destruct H2.
left;apply Union_introl;auto.
destruct H2.
left;apply Union_intror;auto.
right;auto.
destruct H2.
destruct H2.
apply Union_introl;auto.
apply Union_intror;left;auto.
apply Union_intror;right;auto.
Qed.

Lemma subset_finite_eq_rel : forall A:_, eq_dec A -> forall Rsuper:_, Finite A Rsuper ->
  forall R:_, ens_dec _ R -> Included _ R Rsuper -> Finite _ R.
Proof.
intros A eq_dec_A Rsuper H;induction H as [|Rsuper];intros.
rewrite (Extensionality_Ensembles _ R (Empty_set _)).
apply Empty_is_finite.
split;intros;auto.
intro;intros.
inversion H1.

destruct (H1 x).
assert (Finite _ (Subtract _ R x)).
  apply IHFinite;auto.
  intro;intros.
  destruct (eq_dec_A x x0).
  rewrite e in *;right;intro.
  destruct H4;auto.
  apply H5;apply In_singleton.
  destruct (H1 x0).
  left;split;auto.
  intro;apply n;elim H5;auto.
  right;intro.
  destruct H5;auto.
  intro;intros.
  destruct H4.
  destruct (H2 _ H4);auto.
  contradiction.
rewrite (Extensionality_Ensembles _ R (Add _ (Subtract _ R x) x)).
apply Union_is_finite;auto.
intro.
destruct H5.
apply H6;apply In_singleton.
split;intro;intros.
destruct (eq_dec_A x x0).
rewrite e in *;right;apply In_singleton.
left;split;auto.
intro;apply n;elim H6;auto.
destruct H5;destruct H5;auto.

apply IHFinite;auto.
intro;intros.
specialize (H2 _ H4).
destruct H2;auto.
exfalso.
apply H3.
assert (x=x0).
  elim H2;auto.
rewrite H5 in *;auto.
Qed.

Require Import List.

Definition relates_list (A:Type)(e:Ensemble A)(l:list A):=
  forall x:A, (Ensembles.In _ e x) <-> ((List.In) x l).

Lemma relates_list_finite : forall (A:_)(l:_)(R:_),
  eq_dec A -> relates_list _ R l -> Finite A R.
Proof.
induction l;intros.
assert (Same_set _ R (Empty_set _)).
  split;intro;intros.
  destruct (H x).
  specialize (H1 H0).
  inversion H1.
  inversion H0.
rewrite (Extensionality_Ensembles _ _ _ H0);apply Empty_is_finite.

destruct (List.in_dec X a l).
apply IHl;auto.
split;intro;intros.
destruct (H x).
destruct (H1 H0);auto.
rewrite<- H3;auto.
apply H;right;auto.
assert (Finite _ (Subtract _ R a)).
  apply IHl;auto;split;intro;intros.
  destruct H0.
  destruct (H x).
  destruct (H2 H0);auto.
  exfalso;apply H1;rewrite H4;apply In_singleton.
  split.
  apply H;right;auto.
  intro.
  apply n.
  destruct H1;auto.
assert (Same_set _ R (Ensembles.Add _ (Subtract _ R a) a)).
  split;intro;intros.
  destruct (X x a).
  rewrite e;right;apply In_singleton.
  left;split;auto.
  intro;apply n0;destruct H2;auto.
  destruct H1.
  destruct H1;auto.
  apply H;left;destruct H1;auto.
rewrite (Extensionality_Ensembles _ _ _ H1).
apply Union_is_finite;auto.
intro.
destruct H2;apply H3;apply In_singleton.
Qed.

