Add LoadPath projectPath.

Require Import relation_definitions.
Require Import relation_lemmas.
Require Import propositional_formulas.
Require Import relative_consequence.
Require Import BES.

Require Import Finite_sets.
Require Import Ensembles.

(*relations on BES*)
Definition rel_eq_rank (E:_)(R:relation propVar):=
  forall XY:_, R XY -> bnd_block_number E (fst XY) = bnd_block_number E (snd XY).

Definition rel_all_eq_rank (E:_):relation propVar:=
  fun (XY:_) => bnd_block_number E (fst XY) = bnd_block_number E (snd XY).

Definition rel_all_bnd (E:_):relation propVar:=
  fun (XY:_) => bnd E (fst XY) /\ bnd E (snd XY).

Definition rel_all_bnd_eq_rank (E:_):relation propVar:=
  fun (XY:_) => rel_all_eq_rank E XY /\ rel_all_bnd E XY.

Lemma rel_all_eq_rank_bnd_dec : forall E:_, ens_dec _ (rel_all_bnd_eq_rank E).
Proof.
intros;intro;intros.
destruct (bnd_dec E (fst x)).
destruct (bnd_dec E (snd x)).
destruct (eq_dec_nat (bnd_block_number E (fst x)) (bnd_block_number E (snd x))).
left.
split;auto.
split;auto.
right;intro.
apply n.
destruct H;auto.
right;intro.
apply n.
destruct H.
destruct H0;auto.
right;intro.
apply n.
destruct H.
destruct H0;auto.
Qed.

Definition rel_all_bnd_block (b1 b2:_):relation propVar:=
  fun (XY:_) => bnd_block b1 (fst XY) /\ bnd_block b2 (snd XY).

Lemma rel_all_eq_rank_bnd_finite_block : forall b1 b2:_, Finite _ (rel_all_bnd_block b1 b2).
Proof.
induction b1.
intros.
rewrite (Extensionality_Ensembles _ (rel_all_bnd_block nil b2) (Empty_set _)).
apply Empty_is_finite.
split;intro;intros;inversion H.
inversion H0.

intro.
assert (Finite _ (rel_all_bnd_block (a::nil) b2)).
  induction b2.
  rewrite (Extensionality_Ensembles _ (rel_all_bnd_block (a::nil) nil) (Empty_set _)).
  apply Empty_is_finite.
  split;intro;intros;inversion H.
  inversion H1.
  destruct a as [x fx].
  destruct a0 as [y fy].
  rewrite (Extensionality_Ensembles _ (rel_all_bnd_block (bEqn x fx :: nil) (bEqn y fy :: b2)) (Add _ (rel_all_bnd_block (bEqn x fx :: nil) b2) (x,y))).
  apply (Add_is_finite _ (eq_dec_pairs _ eq_dec_nat));auto.
  split;intro;intros.
  destruct H;simpl in *.
  destruct H;try contradiction.
  destruct x0 as (x1,y1);simpl in *.
  destruct (eq_dec_nat y1 y).
  rewrite e in *;rewrite H in *;right;apply In_singleton.
  destruct H0;try contradiction.
  left;split;simpl;auto.
  destruct H.
  destruct H.
  split;auto.
  destruct x0;simpl in *;auto.
  destruct x0 as (x0,y0).
  assert ((x,y)=(x0,y0)).
    elim H;auto.
  assert (x=x0).
    congruence.
  assert (y=y0).
    congruence.
  split;simpl;auto.
rewrite (Extensionality_Ensembles _ (rel_all_bnd_block (a :: b1) b2) (Union _ (rel_all_bnd_block (a :: nil) b2) (rel_all_bnd_block b1 b2))).
apply finite_union;auto.
apply eq_dec_pairs;apply eq_dec_nat.
destruct a as [z fz].
split;intro;intros.
destruct x as (x,y).
destruct H0;simpl in *.
destruct H0.
apply Union_introl;split;simpl;auto.
apply Union_intror;split;auto.
destruct H0;destruct x as (x,y);destruct H0;split;simpl in *;auto.
destruct H0;try contradiction;auto.
Qed.

Lemma rel_all_eq_rank_bnd_finite : forall E:_, (w_d E) -> Finite _ (rel_all_bnd_eq_rank E).
Proof.
induction E.
intro w_d_E.
rewrite (Extensionality_Ensembles _ (rel_all_bnd_eq_rank nil) (Empty_set _)).
apply Empty_is_finite.
split;intro;intros.
destruct H.
destruct H0;inversion H0.
inversion H.

intro w_d_E.
specialize (IHE (w_d_shorten _ _ w_d_E)).
rewrite (Extensionality_Ensembles _ (rel_all_bnd_eq_rank (a :: E)) (Union _ (rel_all_bnd_eq_rank E) (rel_all_bnd_block a a))).
apply finite_union;auto.
apply (eq_dec_pairs _ eq_dec_nat).
apply rel_all_eq_rank_bnd_finite_block.
split;intro;intros.
destruct a as [b ne_b].
destruct H.
destruct H0.
unfold rel_all_eq_rank in H;simpl in *.
destruct (bnd_block_dec b (fst x));destruct (bnd_block_dec b (snd x));inversion H.
apply Union_intror;split;auto.
destruct H0;try contradiction;destruct H1;try contradiction.
apply Union_introl;split;auto.
split;auto.
destruct H.
destruct H.
split.
unfold rel_all_eq_rank;unfold rel_all_eq_rank in H.
simpl;rewrite H.
destruct H0.
specialize (w_d_bnd_list _ _ w_d_E _ H0).
specialize (w_d_bnd_list _ _ w_d_E _ H1);intros.
destruct (bnd_block_dec a (fst x));try contradiction.
destruct (bnd_block_dec a (snd x));try contradiction;auto.
split;destruct H0;simpl;auto.
destruct H.
split.
unfold rel_all_eq_rank.
simpl.
destruct (bnd_block_dec a (fst x));try contradiction.
destruct (bnd_block_dec a (snd x));try contradiction;auto.
split;simpl;auto.
Qed.

Definition relates_bound (E:_)(R:_):=
  forall XY:_, R XY -> (bnd E (fst XY) /\ bnd E (snd XY)).

Definition rhs_cons (E:_)(R:relation propVar):=
  forall XY:_, R XY -> rel_cons R (rhs E (fst XY)) (rhs E (snd XY)).

