Add LoadPath projectPath.

Require Import relation_definitions.
Require Import relation_lemmas.
Require Import propositional_formulas.
Require Import propositional_formulas_semantics.

Require Import Ensembles.
Require Import Bool.

(*----------rel cons----------*)

Definition consistent_environment (theta : environment)(R:relation propVar) := 
  forall x y: propVar, (R (x,y)) -> (theta x -> theta y).

Lemma environment_union_consistent : forall R:relation propVar, forall th1 th2:environment,
consistent_environment th1 R -> consistent_environment th2 R -> consistent_environment (environment_union th1 th2) R.
Proof.
intros.
intro;intros.
apply orb_true_iff in H2.
destruct H2.
apply orb_true_iff.
left.
apply (H x);auto.
apply orb_true_iff.
right.
apply (H0 x);auto.
Qed.

Lemma full_environment_consistent : forall R:relation propVar, consistent_environment full_environment R.
Proof.
intros;intro;intros.
unfold full_environment;auto.
Qed.

Lemma empty_environment_consistent : forall R:relation propVar, consistent_environment empty_environment R.
Proof.
intros;intro;intros.
inversion H0.
Qed.

Definition rel_cons (R:relation propVar)(f g:propForm):=
forall theta:environment, consistent_environment theta R -> [[f]]theta -> [[g]]theta.

Lemma consistent_environment_rel_union : forall theta:environment, forall R1 R2:relation propVar,
consistent_environment theta (Union _ R1 R2)
->
(consistent_environment theta R1 /\ consistent_environment theta R2).
Proof.
intros.
unfold consistent_environment in *.
split.
intros.
apply (H x y).
left;auto.
assumption.
intros.
apply (H x y).
right;auto.
assumption.
Qed.

Lemma rel_union_consistent_environment : forall theta:environment, forall R1 R2:relation propVar,
(consistent_environment theta R1 /\ consistent_environment theta R2)
->
consistent_environment theta (Union _ R1 R2).
Proof.
intros.
unfold consistent_environment.
destruct H.
intros.
remember (x,y) as XY.
induction H1;rewrite HeqXY in *.
apply (H x);auto.
apply (H0 x);auto.
Qed.




