Add LoadPath projectPath.

Require Import relation_definitions.
Require Import relation_lemmas.
Require Import propositional_formulas.
Require Import propositional_formulas_semantics.
Require Import consequence_propositional_formula.
Require Import propositional_formula_DNF.
Require Import transitive_closure_definitions.
Require Import transitive_closure_minimal_environment.
Require Import relative_consequence.

Require Import Finite_sets.
Require Import Ensembles.
Require Import Bool.

(*reachable*)

Lemma reachable_from_consistent : forall (R:relation propVar)(l:_)
  (H:relates_list _ R l)(x:propVar),
    consistent_environment (reachable_from _ eq_dec_nat R l H x) R.
Proof.
intros;intro;intros.
unfold reachable_from in *.
destruct (reachable_dec_b nat eq_dec_nat l R H x x0).
destruct (reachable_dec_b nat eq_dec_nat l R H x y);auto.
exfalso.
apply n.
apply (path_reachable _ _ _ x0);auto.
destruct (eq_dec_nat x x0);try inversion H1.
rewrite e in *.
destruct (reachable_dec_b nat eq_dec_nat l R H x0 y);auto.
exfalso.
apply n0.
apply next_reachable;auto.
Qed.

Lemma reachable_var_rel_cons_var : forall R:relation propVar, forall H:Finite _ R,
  forall XY:_, reachable nat R XY -> rel_cons R (var (fst XY)) (var (snd XY)).
Proof.
intros.
induction H0;intro;intros.
apply (H1 x);auto.
simpl in *.
apply (H2 y);auto.
apply IHreachable;auto.
Qed.

Lemma reachable_from_self : forall (R:relation propVar)(l:_)
  (H:relates_list _ R l)(x:propVar), reachable_from _ eq_dec_nat R l H x x.
Proof.
intros.
unfold reachable_from.
destruct (eq_dec_nat x x).
destruct (reachable_dec_b nat eq_dec_nat l R H x x);auto.
exfalso;apply n;auto.
Qed.

Lemma reachable_from_cons : forall (R:relation propVar)(l:_)
  (H:relates_list _ R l)(x:propVar),
    forall f:propForm, [[f]]reachable_from _ eq_dec_nat R l H x <-> rel_cons R (var x) f.
Proof.
induction f;simpl;split;intros;auto.
intro;intros;auto.
inversion H0.
apply (H0 full_environment);auto.
apply full_environment_consistent.
intro;intros.
unfold reachable_from in H0.
destruct (reachable_dec_b nat eq_dec_nat l R H x p);auto.

apply (reachable_var_rel_cons_var _ (relates_list_finite _ _ _ (eq_dec_pairs _ eq_dec_nat) H) (x,p) r theta);auto.
destruct (eq_dec_nat x p);auto.
rewrite e in *;auto.
inversion H0.
unfold reachable_from.
destruct (reachable_dec_b nat eq_dec_nat l R H x p);auto.
destruct (eq_dec_nat x p);auto.
specialize (H0 (reachable_from _ eq_dec_nat R l H x) (reachable_from_consistent _ l H x) (reachable_from_self _ l H _)).
simpl in H0.
unfold reachable_from in H0.
destruct (reachable_dec_b nat eq_dec_nat l R H x p);try contradiction.
destruct (eq_dec_nat x p);try contradiction;auto.

intro;intros.
simpl;apply orb_true_iff;apply orb_true_iff in H0;destruct H0.
left;apply IHf1;auto.
right;apply IHf2;auto.

specialize (H0 (reachable_from _ eq_dec_nat R l H x) (reachable_from_consistent _ l H x) (reachable_from_self _ l H _)).
simpl in *;apply orb_true_iff;apply orb_true_iff in H0;destruct H0;auto.

intro;intros.
simpl in *;apply andb_true_iff;apply andb_true_iff in H0;destruct H0;split.
apply IHf1;auto.
apply IHf2;auto.

specialize (H0 (reachable_from _ eq_dec_nat R l H x) (reachable_from_consistent _ l H x) (reachable_from_self _ l H _)).
simpl in *;apply andb_true_iff;apply andb_true_iff in H0;destruct H0;split;auto.
Qed.

Fixpoint min_env_for_propForm (R:relation propVar)(l:_)(H:_)(f:propForm):environment:=
match f with
| f1 /\p f2 => environment_union (min_env_for_propForm R l H f1) (min_env_for_propForm R l H f2)
| f1 \/p f2 => environment_union (min_env_for_propForm R l H f1) (min_env_for_propForm R l H f2)
| var x => reachable_from  _ eq_dec_nat R l H x
| _ => empty_environment
end.

Lemma min_env_for_propForm_consistent : forall R:relation propVar, forall l:_, forall H:_, forall f:propForm,
  consistent_environment (min_env_for_propForm R l H f) R.
Proof.
induction f.
intro;intros;auto.
intro;intros;auto.
apply reachable_from_consistent.
simpl.
apply environment_union_consistent;auto.
apply environment_union_consistent;auto.
Qed.

Lemma min_env_for_propForm_self : forall R:relation propVar, forall l:_, forall H:_, forall f:propForm,
  ~propForm_eqv f bot -> [[f]](min_env_for_propForm R l H f).
Proof.
induction f;intros;simpl;auto.
exfalso;apply H0;split;intro;auto.
unfold reachable_from.
destruct (reachable_dec_b nat eq_dec_nat l R H p p);auto.
destruct (eq_dec_nat p p);auto.
destruct (propForm_solution_choice f1 full_environment).
apply orb_true_iff;left;apply environment_union_maintains;apply IHf1.
intro.
assert false.
  apply (H2 full_environment);auto.
inversion H3.
apply orb_true_iff;right;apply environment_union_commutes;apply environment_union_maintains;apply IHf2.
intro.
apply H0.
split;intro.
specialize (full_false_eqv_bot _ H1);intro.
simpl in H3;apply orb_true_iff in H3;destruct H3.
apply H4;auto.
apply H2;auto.
inversion H3.
apply andb_true_iff;split.
apply environment_union_maintains;apply IHf1.
intro.
apply H0.
split;intro.
apply andb_true_iff in H2;destruct H2;apply H1;auto.
inversion H2.
apply environment_union_commutes;apply environment_union_maintains;apply IHf2.
intro.
apply H0.
split;intro.
apply andb_true_iff in H2;destruct H2;apply H1;auto.
inversion H2.
Qed.

Lemma consistent_env_reachable : forall R:relation propVar, forall H:Finite _ R, 
  forall theta:environment, consistent_environment theta R -> forall XY:_,
    reachable _ R XY -> theta (fst XY) -> theta (snd XY).
Proof.
intros.
induction H1;simpl in *.
apply (H0 x);auto.
apply (H0 y);auto.
Qed.

Lemma min_env_conj : forall (g:propForm)(R:relation propVar)(l:_)(H:_)(f:propForm),
  conj_propForm f -> ~propForm_eqv f bot -> rel_cons R f g <-> [[g]]min_env_for_propForm R l H f.
Proof.
induction g;split;intros;auto.
intro;intros;auto.
apply H2.
apply min_env_for_propForm_consistent.
apply min_env_for_propForm_self;auto.
inversion H2.

apply H2.
apply min_env_for_propForm_consistent.
apply min_env_for_propForm_self;auto.

intro;intros.
induction H0 using conj_ind.
  inversion H2.
  inversion H4.
  simpl in H2.
  unfold reachable_from in H2.
  destruct (eq_dec_nat x p).
  rewrite e in *;auto.
  destruct (reachable_dec_b _ eq_dec_nat l R H x p).
  apply (consistent_env_reachable _ (relates_list_finite _ _ _ (eq_dec_pairs _ eq_dec_nat) H) _ H3 (x,p));auto.
  inversion H2.
simpl in H2.
simpl in H4;apply andb_true_iff in H4;destruct H4.
apply orb_true_iff in H2;destruct H2.
apply IHconj_propForm1;auto.
intro.
apply H1.
split;intro.
simpl in H6;apply andb_true_iff in H6;destruct H6;apply H5;auto.
inversion H6.
apply IHconj_propForm2;auto.
intro.
apply H1.
split;intro.
simpl in H6;apply andb_true_iff in H6;destruct H6;apply H5;auto.
inversion H6.
apply H2.
apply min_env_for_propForm_consistent.
apply min_env_for_propForm_self;auto.
intro;intros.
simpl in *;apply orb_true_iff;apply orb_true_iff in H2;destruct H2.
destruct (IHg1 _ _ H _ H0 H1).
left;apply H6;auto.
destruct (IHg2 _ _ H _ H0 H1).
right;apply H6;auto.
apply H2.
apply min_env_for_propForm_consistent.
apply min_env_for_propForm_self;auto.
intro;intros.
simpl in *;apply andb_true_iff;apply andb_true_iff in H2;destruct H2.
split.
apply (IHg1 _ _ H _ H0 H1);auto.
apply (IHg2 _ _ H _ H0 H1);auto.
Qed.

