Add LoadPath projectPath.

Require Import Finite_sets.
Require Import Ensembles.
Require Import relation_definitions.
Require Import relation_lemmas.
Require Import propositional_formulas.
Require Import propositional_formulas_semantics.
Require Import transitive_closure_definitions.
Require Import relative_consequence.
Require Import BES.
Require Import relations_on_BES.
Require Import consistent_consequence.

Definition rhs_rel_cons (E:_)(R G:relation propVar):=
  forall XY:_, R XY -> rel_cons (Union _ R G) (rhs E (fst XY)) (rhs E (snd XY)).

Definition relative_cc (E:_)(R G:relation propVar):Prop :=
  rel_eq_rank E R /\ rhs_rel_cons E R G.

Lemma empty_relation_rel_cc : forall E:_, forall G:relation propVar, 
  relative_cc E (Empty_set _) G.
Proof.
intros.
split;intro;intros;inversion H.
Qed.

Lemma empty_relative_cc: forall E:_, forall R:_,
  relative_cc E R (Empty_set _) -> consistent_consequence E R.
Proof.
intros.
destruct H.
split;intro;intros;auto.
intro;intros.
apply H0;auto.
intro;intros.
apply (H2 x);auto.
remember (x,y) as xy.
destruct H4;auto.
inversion H4.
Qed.

Lemma cc_relative_empty: forall E:_, forall R:relation propVar, 
  consistent_consequence E R -> relative_cc E R (Empty_set _).
Proof.
intros.
destruct H.
split;auto.
intro;intros;intro;intros.
apply H0;auto.
apply (consistent_environment_rel_union _ _ _ H2).
Qed.

Lemma rel_union_maintains_rel_cc : forall E:_, forall R1 R2 G:relation propVar, 
  relative_cc E R1 G -> relative_cc E R2 G -> relative_cc E (Union _ R1 R2) G.
Proof.
intros.
destruct H;destruct H0.
split;intro;intros.
destruct H3;auto.
intro;intros.
destruct (consistent_environment_rel_union _ _ _ H4).
destruct (consistent_environment_rel_union _ _ _ H6).
destruct H3.
apply H1;auto.
apply rel_union_consistent_environment;auto.
apply H2;auto.
apply rel_union_consistent_environment;auto.
Qed.

Definition rel_cc_on_propForm (E:_)(G:_)(f g:_):=
  exists R:_, Finite _ R /\ relates_bound E R /\ relative_cc E R G /\ rel_cons (Union (nat * nat) R G) f g.

Lemma rel_cc_trans : forall E:_, forall G:relation propVar, forall a b c:propForm, 
  ((rel_cc_on_propForm E G a b)/\(rel_cc_on_propForm E G b c)) ->
    (rel_cc_on_propForm E G a c).
Proof.
intros.
destruct H.
destruct H as [R1 H].
destruct H.
destruct H1.
destruct H0 as [R2 H0].
destruct H0.
destruct H3.
exists (Union _ R1 R2).
split.
apply finite_union;auto.
apply (eq_dec_pairs _ eq_dec_nat).
split.
intro;intros.
destruct H5;auto.
destruct H2;destruct H4.
split.
apply rel_union_maintains_rel_cc;auto.
intro;intros.
destruct (consistent_environment_rel_union _ _ _ H7).
destruct (consistent_environment_rel_union _ _ _ H9).
apply H6.
apply rel_union_consistent_environment;auto.
apply H5;auto.
apply rel_union_consistent_environment;auto.
Qed.

(*----------rel_cc----------*)

Definition rel_cc_max (E:_)(G:_):=
  fun (XY:_) => exists R:_, rel_eq_rank E R /\ relates_bound E R /\ rhs_rel_cons E R G /\ (Union _ R G) XY.

Lemma rel_cc_max_rel_all_eq_rank_bnd :
  forall E:_, Same_set _ (rel_cc_max E (rel_all_bnd_eq_rank E)) (rel_all_bnd_eq_rank E).
Proof.
split;intro;intros.
destruct H as [R H].
destruct H.
destruct H0.
destruct H1.
destruct H2;auto.
destruct (H0 _ H2).
specialize (H _ H2).
split;auto.
split;auto.

exists (Empty_set _).
split.
intro;intros;inversion H0.
split.
intro;intros;inversion H0.
split.
intro;intros;inversion H0.
apply Union_intror;auto.
Qed.

Lemma rel_cc_max_included : forall (E:_)(G:_), Included _ (rel_cc_max E G) (Union _ G (rel_all_bnd_eq_rank E)).
Proof.
intros;intro;intros.
destruct H as [R H].
destruct H.
destruct H0.
destruct H1.
destruct H2;auto.
apply Union_intror.
destruct (H0 _ H2).
specialize (H _ H2).
split;auto.
split;auto.
apply Union_introl;auto.
Qed.

Lemma cc_rel_max_cons_env_has_cc : forall (E:_)(G:_)(theta:_),
  consistent_environment theta (rel_cc_max E G) -> forall XY:_, (rel_cc_max E G XY) ->
    exists R:_, relative_cc E R G /\ (Union _ R G) XY /\
      consistent_environment theta (Union _ R G).
Proof.
intros.
destruct H0 as [R H0].
destruct H0.
destruct H1.
destruct H2.
exists R.
split;auto.
split;auto.
split;auto.
intro;intros.
apply (H x);auto.
exists R;auto.
Qed.

Lemma rel_cc_max_cc_max : forall E:_,
  Included _ (rel_cc_max E (Empty_set _)) (cc_max E).
Proof.
intros;intro;intros.
destruct H as [R H].
exists R.
destruct H;destruct H0.
destruct H1.
destruct H2;try contradiction.
split;auto.
split;auto.
intro;intros;intro;intros.
apply (H1 XY);auto.
intro;intros.
apply (H4 x0);auto.
destruct H6;auto;contradiction.
Qed.

Lemma cc_max_rel_cc_max : forall E:_,
  Included _ (cc_max E) (rel_cc_max E (Empty_set _)).
Proof.
intros;intro;intros.
destruct H as [R H].
destruct H;destruct H;destruct H0.
exists R.
split;auto.
split;auto.
split.
intro;intros;intro;intros.
apply (H1 XY);auto.
intro;intros.
apply (H4 x0);auto.
apply Union_introl;auto.
apply Union_introl;auto.
Qed.

Lemma eq_cc_max_rel_cc_max : forall E:_,
  (cc_max E)=(rel_cc_max E (Empty_set _)).
Proof.
intros;apply Extensionality_Ensembles.
split.
apply cc_max_rel_cc_max.
apply rel_cc_max_cc_max.
Qed.

Lemma rel_cc_reach : forall E:_, forall R:_, relative_cc E R (Empty_set _) -> relative_cc E (reachable _ R) (Empty_set _).
Proof.
intros;split.
intro;intros.
induction H0.
destruct H.
apply H;auto.
destruct H.
specialize (H _ H1).
simpl in *;rewrite IHreachable;auto.

destruct H.
intro;intros.
induction H1.
intro;intros.
apply (H0 _ H1 theta);auto.
destruct (consistent_environment_rel_union _ _ _ H2).
apply rel_union_consistent_environment;split;auto.
intro;intros.
apply (H4 x0);auto.
apply next_reachable;auto.

intro;intros.
simpl.
specialize (H0 _ H2 theta);simpl in *;apply H0;auto.
intro;intros.
apply (H3 x0);auto.
remember (x0,y0) as xy.
apply Union_introl;destruct H5;destruct x1.
apply next_reachable;auto.
inversion H5.
Qed.

















