Add LoadPath projectPath.

Require Import relation_definitions.
Require Import relation_lemmas.
Require Import Ensembles.
Require Import Bool.

(*TODO rename reachable to trans_close*)
(*--------reachable--------*)

Inductive reachable (A:Type)(R:relation A):relation A:=
| next_reachable : forall x y:A, R (pair x y) -> reachable A R (pair x y)
| path_reachable : forall x y:A, reachable A R (pair x y) -> forall z:A, R (pair y z) -> reachable A R (pair x z).

Lemma add_maintains_reachable : forall A:Type, forall R:relation A, forall x y:A,
  reachable A R (x,y) -> forall a b:A, reachable A (Add (A*A) R (a,b)) (x,y).
Proof.
intros;induction H.
apply next_reachable;left;auto.

apply (path_reachable _ _ _ y0);auto;left;auto.
Qed.

Lemma reachable_trans : forall A:Type, forall R:relation A,
  forall YZ:A*A, reachable A R YZ ->
    forall x:A, reachable A R (pair x (fst YZ)) -> reachable A R (pair x (snd YZ)).
Proof.
intros A R YZ H.
induction H.
intros.
simpl in *.
apply (path_reachable _ _ _ x);auto.

simpl in *.
intros.
apply (path_reachable _ _ _ y);auto.
Qed.

Lemma trans_reachable : forall A:Type, forall R:relation A,
  forall XZ:A*A, reachable A R XZ ->
    (In (A*A) R XZ \/ (exists y:A, (reachable A R (fst XZ,y)) /\ reachable A R (y,snd XZ))).
Proof.
intros.
induction H;simpl.
left;auto.

destruct IHreachable.
right;exists y;split;auto.
apply next_reachable;auto.

destruct H1.
destruct H1.
simpl in *.
right;exists x0;split;auto.
apply (path_reachable _ _ _ y);auto.
Qed.

Lemma reachable_add : forall A:Type, eq_dec A -> forall R:relation A, forall XY:A*A,
  reachable A R XY -> forall x0 y0:A, In (A*A) R (x0,y0) ->
    (reachable A (Subtract (A*A) R (x0,y0)) XY \/
    ((reachable A (Subtract (A*A) R (x0,y0)) (fst XY,x0) \/ x0=fst XY)
     /\ (reachable A (Subtract (A*A) R (x0,y0)) (y0,snd XY) \/ y0=snd XY))).
Proof.
intros A H R XY H0.
induction H0 as [|x y H0 IH z];intros.
destruct (eq_dec_pairs _ H (x0,y0) (x,y)).
right;split;right;simpl;congruence.
left;apply next_reachable.
split;auto.
intro.
apply n.
elim H2;auto.

destruct (eq_dec_pairs _ H (x0,y0) (x,z)).
right;split;right;simpl;congruence.
destruct (IH x0 y0);auto.
destruct (H x0 x).
rewrite e in *.
simpl.
destruct (H y x).
rewrite e0 in *.
left;apply next_reachable;auto.
split;auto.
intro.
apply n.
elim H4;auto.
left.
apply (path_reachable _ _ _ y);auto.
split;auto.
intro.
apply n0.
assert ((x,y0)=(y,z)).
  elim H4;auto.
congruence.
simpl in *.
destruct (H y0 z).
rewrite e in *.
destruct (H x0 y).
rewrite e0 in *.
right;split;auto.
left.
apply (path_reachable _ _ _ y);auto.
split;auto.
intro.
apply n1.
assert ((x0,z)=(y,z)).
  elim H4;auto.
congruence.
left.
apply (path_reachable _ _ _ y);auto.
split;auto.
intro.
apply n1.
assert ((x0,y0)=(y,z)).
  elim H4;auto.
congruence.
destruct H3.
simpl in *.
destruct H4.
destruct (H y0 z);auto.
right;split;auto;left.
apply (path_reachable _ _ _ y);auto.
split;auto.
intro.
apply n0.
assert ((x0,y0)=(y,z)).
  elim H5;auto.
congruence.
rewrite H4 in *.
destruct (H y z);auto.
right;split;auto;left.
apply next_reachable.
split;auto.
intro.
apply n0.
assert ((x0,y)=(y,z)).
  elim H5;auto.
congruence.
Qed.

Lemma reachable_eqv : forall A:Type, forall R1:relation A, forall x y:A,
  reachable A R1 (x,y) -> forall R2:relation A, Same_set (A*A) R1 R2 ->
    reachable A R2 (x,y).
Proof.
intros.
induction H.
apply next_reachable.
apply H0;auto.

apply (path_reachable _ _ _ y0);auto.
apply H0;auto.
Qed.

