Add LoadPath projectPath.

Require Import relation_definitions.
Require Import relation_lemmas.
Require Import transitive_closure_definitions.
Require Import Ensembles.
Require Import List.
Require Import Finite_sets.

Definition relates_list (A:Type)(e:Ensemble A)(l:list A):=
  forall x:A, (Ensembles.In _ e x) <-> ((List.In) x l).

Lemma relates_list_finite : forall (A:_)(l:_)(R:_),
  eq_dec A -> relates_list _ R l -> Finite A R.
Proof.
induction l;intros.
assert (Same_set _ R (Empty_set _)).
  split;intro;intros.
  destruct (H x).
  specialize (H1 H0).
  inversion H1.
  inversion H0.
rewrite (Extensionality_Ensembles _ _ _ H0);apply Empty_is_finite.

destruct (List.in_dec X a l).
apply IHl;auto.
split;intro;intros.
destruct (H x).
destruct (H1 H0);auto.
rewrite<- H3;auto.
apply H;right;auto.
assert (Finite _ (Subtract _ R a)).
  apply IHl;auto.
  split;intro;intros.
  destruct H0.
  destruct (H x).
  destruct (H2 H0);auto.
  exfalso;apply H1;rewrite H4;apply In_singleton.
  split.
  apply H;right;auto.
  intro.
  apply n.
  destruct H1;auto.
assert (Same_set _ R (Ensembles.Add _ (Ensembles.Subtract _ R a) a)).
  split;intro;intros.
  destruct (X x a).
  rewrite e;right;apply In_singleton.
  left;split;auto.
  intro;apply n0;destruct H2;auto.
  destruct H1.
  destruct H1;auto.
  apply H;left;destruct H1;auto.
rewrite (Extensionality_Ensembles _ _ _ H1).
apply Union_is_finite;auto.
intro.
destruct H2;apply H3;apply In_singleton.
Qed.

Lemma relates_list_dec (A:Type): eq_dec A -> forall l:list A, forall e:Ensemble A,
  relates_list _ e l -> (forall x:_, {Ensembles.In _ e x} + {~Ensembles.In _ e x}).
Proof.
induction l;intros.
right;intro.
destruct (H x).
apply (H1 H0).

destruct (X a x).
left.
apply H;rewrite e0;simpl;auto.
destruct (in_dec X x l).
left.
apply H;simpl;auto.
right;intro.
destruct (H x).
destruct (H1 H0);auto.
Qed.

Lemma exists_list : forall A:Type, forall e:Ensemble A, Finite A e ->
  exists l:list A, (forall x:_, Ensembles.In _ e x <-> List.In x l).
Proof.
intros.
induction H.
exists nil;split;intro;inversion H.

destruct IHFinite as [l IH].
exists (x::l).
split;intros.
destruct H1.
right;apply IH;auto.
left;destruct H1;auto.
destruct H1.
right;rewrite H1;apply In_singleton.
left;apply IH;auto.
Qed.

Lemma reachable_trans2 : forall A:_, forall R:_,
  forall YZ:_, reachable A R YZ -> forall XY:_, reachable _ R XY ->
    (snd XY=fst YZ) -> reachable _ R (fst XY,snd YZ).
Proof.
intros A R YZ H;induction H as [y1 z H|y1 z0 H IH z1 H0];intros;destruct XY as (x,y0);simpl in *.
rewrite<- H1 in *.
apply (path_reachable _ _ _ y0);auto.

rewrite<- H2 in *.
apply (path_reachable _ _ _ z0);auto.
apply (IH _ H1);auto.
Qed.

Lemma reachable_dec_b : forall (A:_)(X:eq_dec A)(l:_)(R:relation A),
  relates_list _ R l -> forall x y:A, {reachable A R (pair x y)} + {~(reachable A R (pair x y))}.
Proof.
induction l;intros.
right;intro.
induction H0;auto.
apply (H (x0,y0));auto.
destruct (in_dec (eq_dec_pairs _ X) a l).
apply IHl.
split;intro;intros.
destruct (H x0).
destruct (H1 H0);auto.
rewrite<- H3;auto.
apply H;right;auto.
assert (relates_list _ (Subtract _ R a) l).
  split;intro;intros.
  destruct H0.
  destruct (H x0).
  destruct (H2 H0);auto.
  exfalso;apply H1;rewrite H4;apply In_singleton.
  split.
  apply H;right;auto.
  intro.
  apply n.
  destruct H1;auto.
destruct a as [a b].
assert (Same_set _ R (Ensembles.Add _ (Subtract _ R (a,b)) (a,b))).
  split;intro;intros.
  destruct (eq_dec_pairs _ X x0 (a,b)).
  rewrite e in *.
  right;apply In_singleton.
  left;split;auto.
  intro.
  apply n0;elim H2;auto.
  destruct H1;auto.
  destruct H1;auto.
  apply H;elim H1;left;auto.
destruct (IHl _ H0 x y).
left.
rewrite (Extensionality_Ensembles _ _ _ H1).
apply add_maintains_reachable;auto.
destruct (IHl _ H0 x a).
destruct (IHl _ H0 b y).
left.
apply (reachable_trans _ _ (b,y)).
rewrite (Extensionality_Ensembles _ _ _ H1).
apply add_maintains_reachable;auto.
simpl.
apply (path_reachable _ _ _ a);auto.
rewrite (Extensionality_Ensembles _ _ _ H1).
apply add_maintains_reachable;auto.
apply H;left;auto.
destruct (X y b).
rewrite e in *.
left.
apply (path_reachable _ _ _ a).
rewrite (Extensionality_Ensembles _ _ _ H1).
apply add_maintains_reachable;auto.
apply H;left;auto.
right;intro.
destruct (reachable_add _ X _ _ H2 a b);auto.
apply H;left;auto.
destruct H3.
simpl in *.
destruct H4;auto.
destruct (IHl _ H0 b y).
destruct (X a x).
left.
apply (reachable_trans _ _ (b,y)).
rewrite (Extensionality_Ensembles _ _ _ H1).
apply add_maintains_reachable;auto.
simpl.
apply next_reachable.
rewrite<- e.
apply H;left;auto.
right;intro.
destruct (reachable_add _ X _ _ H2 a b);auto.
apply H;left;auto.
destruct H3.
simpl in *.
destruct H3;auto.
destruct (eq_dec_pairs _ X (a,b) (x,y)).
left.
apply next_reachable.
apply H;rewrite e;left;auto.
right;intro.
destruct (reachable_add _ X _ _ H2 a b);auto.
apply H;left;auto.
destruct H3.
simpl in *.
destruct H4;auto.
destruct H3;auto.
apply n3.
congruence.
Qed.

Definition reachable_from (A:Type)(eqA:eq_dec A)(R:relation A)(l:_)(H:relates_list _ R l)
  (x:A):A -> bool:=
fun (y:A) => if reachable_dec_b A eqA l R H x y then true else if eqA x y then true else false.